from musurgia.pdf.segmentedline import SegmentedLine
from musurgia.tree import Tree


class AbstractVoice(SegmentedLine):
    def __init__(self, length, unit=10, *args, **kwargs):
        self._length = None
        self.length = length
        self._unit = None
        self.unit = unit
        super().__init__(lengths=self.length * [self.unit], *args, **kwargs)

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, val):
        if not isinstance(val, int):
            raise TypeError('length {} must be of type int not {}'.format(val, type(val)))
        if val < 1:
            raise ValueError('length {} must be a positive non zero integer'.format(val))
        self._length = val

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, val):
        if val < 0:
            raise ValueError('unit {} must be a positive number'.format(val))
        self._unit = val

    def update_unit(self, unit):
        if self.unit == unit:
            pass
        else:
            self.unit = unit
            for line in self.line_segments:
                line.length = unit



import os
from itertools import cycle
from math import ceil

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from quicktions import Fraction

from musurgia.unittest import TestCase
from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.basic_functions import slice_list
from musurgia.chordfield.chordfield import ChordField, Breathe
from musurgia.chordfield.valuegenerator import ValueGenerator
from musurgia.fractaltree.fractalmusic import FractalMusic
from musurgia.interpolation import Interpolation, RandomInterpolation

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        fm = FractalMusic(quarter_duration=20, tempo=80)
        fm.midi_generator.midi_range = [60, 84]
        fm.add_layer()
        sorted_children = sorted(fm.get_children(), key=lambda child: child.fractal_order)
        chord_field = ChordField(
            quarter_duration=10,
            duration_generator=ValueGenerator(ArithmeticProgression(a1=0.2, an=2)),
            midi_generator=ValueGenerator(Interpolation(start=84, end=60,
                                                        key=lambda midi: round(midi * 2) / 2)),
            short_ending_mode='add_rest'
        )
        sorted_children[-1].chord_field = chord_field
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_1.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_2(self):
        def add_chord_field(child):
            child.chord_field = ChordField(duration_generator=ValueGenerator(ArithmeticProgression(a1=0.2, an=2)),
                                           midi_generator=ValueGenerator(
                                                Interpolation(start=child.midi_generator.midi_range[0],
                                                              end=child.midi_generator.midi_range[1],
                                                              duration=None,
                                                              key=lambda
                                                                  midi: round(midi * 2) / 2)),
                                           short_ending_mode='stretch')

        fm = FractalMusic(quarter_duration=20, tempo=80, proportions=[1, 2, 3, 4, 5],
                          tree_permutation_order=[3, 1, 5, 2, 4])
        fm.midi_generator.midi_range = [60, 84]
        fm.add_layer()
        sorted_children = sorted(fm.get_children(), key=lambda child: child.fractal_order)
        add_chord_field(sorted_children[-1])

        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + '_test_2.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_3(self):
        # node.chord_fields are part of a group
        cfg = ChordField(
            duration_generator=ValueGenerator(RandomInterpolation(start=[0.25, 0.25], end=[0.75, 1], seed=20)))
        cf_1 = ChordField(
            midi_generator=ValueGenerator(cycle([60])))
        cf_2 = ChordField(
            midi_generator=ValueGenerator(cycle([72])),
            long_ending_mode='cut')

        cfg.add_child(cf_1)
        cfg.add_child(cf_2)

        fm = FractalMusic(quarter_duration=20, tempo=80)
        fm.add_layer()
        fm.get_children()[0].chord_field = cf_1
        fm.get_children()[1].chord_field = cf_2
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + 'test_3.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_4(self):
        cf_1 = ChordField(
            quarter_duration=10,
            midi_generator=ValueGenerator(cycle([60, 61, 64, 66])),
            long_ending_mode='self_extend',
            short_ending_mode='self_shrink'
        )
        cf_2 = ChordField(
            quarter_duration=3,
            midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])),
            long_ending_mode='self_extend',
            short_ending_mode='self_shrink'
        )

        breathe_unit = Fraction(1, 5)
        breathe_breakpoints = (5 * breathe_unit, breathe_unit, 5 * breathe_unit)
        breathe_proportions = [2, 4, 1, 7, 2]

        breathe = Breathe(proportions=breathe_proportions,
                          quarter_duration=13,
                          breakpoints=breathe_breakpoints)
        cfg = ChordField(duration_generator=breathe.duration_generator)
        cfg.add_child(cf_1)
        cfg.add_child(cf_2)

        simple_format = SimpleFormat()
        simple_format.extend(cf_1.simple_format)
        simple_format.extend(cf_2.simple_format)

        self.score.set_time_signatures(ceil(simple_format.quarter_duration))
        simple_format.to_stream_voice().add_to_score(self.score)
        xml_path = path + 'test_4.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_5(self):
        # node.chord_fields are part of a breathe group
        fm = FractalMusic(quarter_duration=20, tempo=80)
        fm.add_layer()
        fm.quantize_children(grid_size=1)

        node_groups = slice_list(fm.get_children(), (2, 1))

        cf_1 = ChordField(
            midi_generator=ValueGenerator(cycle([60, 61, 64, 66])),
            long_ending_mode='self_extend',
            short_ending_mode='self_shrink'
        )
        cf_2 = ChordField(
            midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])),
            long_ending_mode='self_extend',
            short_ending_mode='self_shrink'
        )

        breathe_unit = Fraction(1, 5)
        breathe_breakpoints = (5 * breathe_unit, breathe_unit, 5 * breathe_unit)
        breathe_proportions = [2, 4, 1, 7, 2]

        breathe = Breathe(proportions=breathe_proportions,
                          quarter_duration=sum([node.chord.quarter_duration for node in node_groups[0]]),
                          breakpoints=breathe_breakpoints)

        cfg = ChordField(duration_generator=breathe.duration_generator)
        cfg.add_child(cf_1)
        cfg.add_child(cf_2)

        fm.get_children()[0].chord_field = cf_1
        fm.get_children()[1].chord_field = cf_2
        score = fm.get_score(show_fractal_orders=True)
        xml_path = path + 'test_5.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)



import os
from unittest import TestCase

from musurgia.fractaltree.fractaltree import FractalTree

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def test_1(self):
        ft = FractalTree(proportions=(1, 2, 3, 4, 5), tree_permutation_order=(3, 5, 1, 2, 4), value=10)
        ft.add_layer()
        # ft.add_layer()
        # print(ft.get_leaves(key=lambda leaf: leaf.index))
        # print(ft.get_leaves(key=lambda leaf: leaf.fractal_order))
        # print(ft.get_leaves(key=lambda leaf: round(float(leaf.value), 2)))
        ft.merge_children(1, 2, 2)
        # print(ft.get_leaves(key=lambda leaf: leaf.index))
        self.assertEqual(ft.get_leaves(key=lambda leaf: leaf.fractal_order), [3, 5, 2])
        self.assertEqual(ft.get_leaves(key=lambda leaf: round(float(leaf.value), 2)), [2.0, 4.0, 4.0])


from unittest import TestCase

from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.quantize import get_quantized_values


class Test(TestCase):
    def setUp(self) -> None:
        self.arith_prog = ArithmeticProgression()

    def test_1(self):
        result = {'a1': None, 'an': None, 'n': None, 'd': None, 's': None}
        self.assertEqual(result, self.arith_prog._parameters_dict)

    def test_2(self):
        self.arith_prog.n = 15
        self.arith_prog.a1 = 1
        self.arith_prog.d = 2
        result = {'a1': 1, 'an': 29, 'n': 15, 'd': 2, 's': 225.0}
        self.assertEqual(result, self.arith_prog.parameters_dict)

    def test_3(self):
        self.arith_prog.n = 15
        self.arith_prog.a1 = 1
        self.arith_prog.d = 2
        result = [1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 27, 29]
        self.assertEqual(result, list(self.arith_prog))

    def test_4(self):
        self.arith_prog.a1 = 36
        self.arith_prog.an = 25
        self.arith_prog.s = 230
        # print(self.arith_prog.parameters_dict)
        values = list(self.arith_prog)
        # print(values)
        quantized = get_quantized_values(values, 1)
        # print(quantized)
        # print(sum(quantized))
        result = [36, 34, 32, 31, 29, 26, 25]
        self.assertEqual(result, quantized)

    def test_5(self):
        self.arith_prog.a1 = 0.2
        self.arith_prog.an = 1.5
        self.arith_prog.s = 20

        self.arith_prog.correct_s = True
        test_case = sum(list(self.arith_prog))
        expected = self.arith_prog.s
        self.assertEqual(expected, test_case)

    def test_6(self):
        self.arith_prog.n = 15
        self.arith_prog.a1 = 1
        self.arith_prog.d = 2
        copy = self.arith_prog.__deepcopy__()
        actual = {'a1': 1, 'an': 29, 'n': 15, 'd': 2, 's': 225.0}
        expected = copy.parameters_dict
        self.assertEqual(expected, actual)

    def test_7(self):
        self.arith_prog.n = 5
        self.arith_prog.a1 = 1
        self.arith_prog.s = 15
        expected = 1
        actual = self.arith_prog.d
        self.assertEqual(expected, actual)

    def test_8(self):
        self.arith_prog.n = 5
        self.arith_prog.a1 = 1
        self.arith_prog.s = 15
        self.arith_prog.d
        expected = 5
        actual = self.arith_prog.an
        self.assertEqual(expected, actual)



from itertools import cycle

from musurgia.chordfield.chordfield import ChordField
from musurgia.chordfield.valuegenerator import ValueGenerator

parent_chord_field = ChordField(duration_generator=ValueGenerator(cycle([0.5])))
parent_chord_field.add_child(ChordField(midi_generator=ValueGenerator(cycle([60]))))
parent_chord_field.add_child(ChordField(quarter_duration=4, midi_generator=ValueGenerator(cycle([60]))))
parent_chord_field.add_child(ChordField(quarter_duration=8, midi_generator=ValueGenerator(cycle([61]))))


print(id(parent_chord_field.duration_generator))
print([id(chfi.duration_generator) for chfi in parent_chord_field.children])
print([chfi._duration_generator for chfi in parent_chord_field.children])

copied_parent_chord_field = parent_chord_field.__deepcopy__()
print(id(copied_parent_chord_field.duration_generator))
print([id(chfi.duration_generator) for chfi in copied_parent_chord_field.children])
print([chfi._duration_generator for chfi in copied_parent_chord_field.children])




from musurgia.pdf.drawobject import DrawObject


class DrawObjectGroup(DrawObject):
    def __init__(self, inner_distance=None, *args, **kwargs):
        self._draw_objects = []
        self._inner_distance = 0
        self.inner_distance = inner_distance
        super().__init__(*args, **kwargs)

    def _set_inner_distances(self):
        if self.draw_objects:
            for draw_object in self.draw_objects[:-1]:
                draw_object.bottom_margin = self.inner_distance
            self.draw_objects[-1].bottom_margin = 0

    def _add_draw_object(self, draw_object):
        if not isinstance(draw_object, DrawObject):
            raise TypeError()

        self._draw_objects.append(draw_object)
        self._set_inner_distances()
        return draw_object

    def add_draw_object(self, draw_object):
        return self._add_draw_object(draw_object)

    @DrawObject.relative_x.setter
    def relative_x(self, val):
        self._relative_x = val
        try:
            for draw_object in self.draw_objects:
                draw_object.relative_x = self.relative_x
        except AttributeError:
            pass

    @DrawObject.relative_y.setter
    def relative_y(self, val):
        self._relative_y = val
        for draw_object in self.draw_objects:
            draw_object.relative_y = self.relative_y

    @property
    def draw_objects(self):
        return self._draw_objects

    @property
    def inner_distance(self):
        return self._inner_distance

    @inner_distance.setter
    def inner_distance(self, val):
        self._inner_distance = val
        self._set_inner_distances()

    def get_relative_x2(self):
        if self.draw_objects:
            return max([do.get_relative_x2() for do in self.draw_objects])
        else:
            return None

    def get_relative_y2(self):
        if self.draw_objects:
            return sum([do.get_relative_y2() + do.bottom_margin for do in self.draw_objects]) + self.relative_y
        else:
            return None

    def draw(self, pdf):
        old_pdf_x = pdf.x
        old_pdf_page = pdf.page
        new_top_margin = 0
        for draw_object in self.draw_objects:
            # print(f'top_margin:{new_top_margin}')
            draw_object.top_margin = new_top_margin
            old_pdf_y = pdf.y
            old_bottom_margin = draw_object.bottom_margin

            draw_object.bottom_margin = self.get_height() + self.bottom_margin - draw_object.get_height()

            pdf.page = old_pdf_page
            pdf.x = old_pdf_x

            draw_object.draw(pdf)

            draw_object.bottom_margin = old_bottom_margin
            new_top_margin = draw_object.get_relative_y2() + draw_object.bottom_margin
            pdf.y = old_pdf_y + draw_object.get_height() + draw_object.bottom_margin

        pdf.y += self.bottom_margin



import os

from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()
        self.fm = FractalMusic(tempo=60, quarter_duration=10)
        self.fm.midi_generator.midi_range = [60, 71]
        self.fm.add_layer()
        self.copied = self.fm.__deepcopy__()

    def test_1(self):
        xml_path = path + '_test_1.xml'
        self.fm.quantize_children(1)
        self.score.set_time_signatures([child.quarter_duration for child in self.fm.get_children()])

        self.copied.inverse_tree_directions()
        self.copied.quantize_children(1)
        self.fm.get_simple_format().to_stream_voice().add_to_score(score=self.score, staff_number=1)
        self.copied.get_simple_format().to_stream_voice().add_to_score(score=self.score, staff_number=2)
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)



import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])

fm = FractalMusic(tempo=72, quarter_duration=10)
fm.midi_generator.midi_range = [60, 72]
fm.add_layer()
node = fm.get_children()[-1]
node.chord.add_grace_chords(TreeChord(midis=node.midi_generator.midi_range[1]), mode='post')
print(node.chord.get_post_grace_chords())
print(fm.get_simple_format().chords[-1].get_post_grace_chords())
score = TreeScoreTimewise()
# fm.get_simple_format().to_stream_voice().add_to_score(score)
simple_format = SimpleFormat(quarter_durations=[1, 4])
simple_format.chords[-1].add_grace_chords(TreeChord(midis=[66]), mode='post')
simple_format.to_stream_voice().add_to_score(score)
# score = fm.get_score(layer_number=fm.number_of_layers)
xml_path = path + '.xml'
score.write(xml_path)



from pathlib import Path

from musurgia.pdf.line import VerticalSegmentedLine
from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase

path = Path(__file__)


class TestVerticalSegmentedLine(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.hsl = VerticalSegmentedLine(lengths=[10, 15, 20, 25])

    def test_draw(self):
        with self.file_path(path, 'draw', 'pdf') as pdf_path:
            self.pdf.translate_page_margins()
            self.pdf.draw_ruler('h')
            self.pdf.draw_ruler('v')
            self.pdf.translate(10, 10)
            self.hsl.draw(self.pdf)
            self.pdf.write(pdf_path)



import os
from fractions import Fraction
from itertools import cycle
from math import ceil

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.random import Random
from musurgia.unittest import TestCase
from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.chordfield.chordfield import ChordField, Breathe
from musurgia.chordfield.valuegenerator import ValueGenerator
from musurgia.interpolation import RandomInterpolation

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        # fields: midi_generators
        # group: no generators
        cfg = ChordField()
        cf_1 = ChordField(quarter_duration=3,
                          midi_generator=ValueGenerator(cycle([60, 61, 64, 66])),
                          duration_generator=ValueGenerator(cycle([1]))
                          )
        cf_2 = ChordField(quarter_duration=6,
                          midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])),
                          duration_generator=ValueGenerator(cycle([1]))
                          )
        cfg.add_child(cf_1)
        cfg.add_child(cf_2)
        xml_path = path + 'test_1.xml'
        cfg.simple_format.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_2(self):
        # fields: midi_generators
        # group: duration_generator with __next__
        cfg = ChordField(
            duration_generator=ValueGenerator(Random(pool=[0.2, 0.4, 0.8, 1.6], seed=10))
        )
        cf_1 = ChordField(quarter_duration=3,
                          midi_generator=ValueGenerator(cycle([60, 61, 64, 66])),
                          long_ending_mode='self_extend'
                          )
        cf_2 = ChordField(quarter_duration=6,
                          midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])),
                          long_ending_mode='self_extend')

        # cfg = ChordFieldGroup(duration_generator=Random(pool=[0.2, 0.4, 0.8, 1.6], seed=10))
        cfg.add_child(cf_1)
        cfg.add_child(cf_2)
        simple_format = cfg.simple_format
        self.score.set_time_signatures(quarter_durations=ceil(cfg.quarter_duration))
        xml_path = path + 'test_2.xml'
        simple_format.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_3(self):
        # fields: both with midi_generators, one duration_generator
        # group: duration_generator with __next__ (Random)
        cfg = ChordField(
            duration_generator=ValueGenerator(Random(pool=[0.2, 0.4, 0.8, 1.6], seed=10))
        )

        cf_1 = ChordField(quarter_duration=3,
                          midi_generator=ValueGenerator(cycle([60, 61, 64, 66])),
                          duration_generator=ValueGenerator(cycle([1]))
                          )
        cf_2 = ChordField(quarter_duration=6,
                          midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])),
                          long_ending_mode='cut'
                          )

        cfg.add_child(cf_1)
        cfg.add_child(cf_2)
        sf_1 = cf_1.simple_format
        sf_2 = cf_2.simple_format
        sf = SimpleFormat()
        sf.extend(sf_1)
        sf.extend(sf_2)
        xml_path = path + 'test_3.xml'
        # cfg.simple_format.to_stream_voice().add_to_score(self.score)
        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_4(self):
        # fields: midi_generators, first one with ending_mode 'post'
        # group: duration_generator with __next__ (Arithmetic Progression)
        cfg = ChordField(
            duration_generator=ValueGenerator(ArithmeticProgression(a1=0.3, an=1.5, correct_s=True))
        )

        cf_1 = ChordField(quarter_duration=3,
                          midi_generator=ValueGenerator(cycle([60, 61, 64, 66])),
                          long_ending_mode='self_extend'
                          )
        cf_2 = ChordField(quarter_duration=6,
                          midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])),
                          short_ending_mode='add_rest',

                          )
        cfg.add_child(cf_1)
        cfg.add_child(cf_2)
        sf = SimpleFormat()
        sf.extend(cf_1.simple_format)
        sf.extend(cf_2.simple_format)
        xml_path = path + 'test_4.xml'
        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_5(self):
        # breathing manually
        times = [3, 7, 3, 10, 3]
        cf_1 = ChordField(quarter_duration=times[0],
                          midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])))
        cf_2 = ChordField(quarter_duration=times[1],
                          midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])))
        cf_3 = ChordField(quarter_duration=times[2],
                          midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])))
        cf_4 = ChordField(quarter_duration=times[3],
                          midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])))
        cf_5 = ChordField(quarter_duration=times[4],
                          midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])))
        points = [1.5, 0.2, 1.5]
        cf_1.duration_generator = ValueGenerator(ArithmeticProgression(a1=points[0], an=points[0], correct_s=True))
        cf_2.duration_generator = ValueGenerator(ArithmeticProgression(a1=points[0], an=points[1], correct_s=True))
        cf_3.duration_generator = ValueGenerator(ArithmeticProgression(a1=points[1], an=points[1], correct_s=True))
        cf_4.duration_generator = ValueGenerator(ArithmeticProgression(a1=points[1], an=points[0], correct_s=True))
        cf_5.duration_generator = ValueGenerator(ArithmeticProgression(a1=points[0], an=points[0], correct_s=True))
        cfg = ChordField()
        cfg.add_child(cf_1)
        cfg.add_child(cf_2)
        cfg.add_child(cf_3)
        cfg.add_child(cf_4)
        cfg.add_child(cf_5)
        #
        # for fractal_tree in cfg.children:
        #     print(fractal_tree.duration_generator.generator.parameters_dict)
        #     values = [float(chord.quarter_duration) for chord in list(fractal_tree)]
        #     print(values)
        #     print(sum(values))

        xml_path = path + 'test_5.xml'
        self.score.set_time_signatures(quarter_durations=times)
        cfg.simple_format.to_stream_voice().add_to_score(self.score, part_number=1)

        self.score.max_division = 5
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_6(self):
        # breathing: duration_generator automatically, midi_generator: InterpolationGroup(2 x RandomInterpolations)
        breathing_proportions = [3, 7, 3, 10, 3]
        breathing_break_points = [1.5, 0.2, 1.5]
        breathing = Breathe(quarter_duration=sum(breathing_proportions), proportions=breathing_proportions,
                            breakpoints=breathing_break_points)

        midi_generator = ValueGenerator()

        midi_generator.add_child(
            ValueGenerator(RandomInterpolation(start=[60, 62, 66, 68], end=[67, 69, 73, 75], seed=10), duration=10)
        )

        midi_generator.add_child(
            ValueGenerator(RandomInterpolation(start=[67, 69, 73, 75], end=[60, 62, 66, 68], seed=11), duration=10)
        )

        breathing.midi_generator = midi_generator
        xml_path = path + 'test_6.xml'
        self.score.set_time_signatures(quarter_durations=breathing_proportions)
        breathing.simple_format.to_stream_voice().add_to_score(self.score)

        self.score.max_division = 5
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_7(self):
        # fields: midi_generators
        # group: duration_generator with __call__ RandomInterpolation
        cfg = ChordField(
            duration_generator=ValueGenerator(
                RandomInterpolation(start=[0.25, 0.25, 0.5], end=[0.5, 0.75, 1], seed=20)))
        cf_1 = ChordField(
            quarter_duration=3,
            midi_generator=ValueGenerator(cycle([60, 61, 64, 66])))
        cf_2 = ChordField(
            quarter_duration=6,
            midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])),
            long_ending_mode='self_extend')

        cfg.add_child(cf_1)
        cfg.add_child(cf_2)
        xml_path = path + 'test_7.xml'
        cfg.simple_format.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_8(self):
        # fields: midi_generators
        # group: duration_generator with __call__ RandomInterpolation
        # output of fields not group

        cfg = ChordField(
            duration_generator=ValueGenerator(
                RandomInterpolation(start=[0.25, 0.25, 0.5], end=[0.5, 0.75, 1], seed=20)))

        cf_1 = ChordField(
            quarter_duration=3,
            midi_generator=ValueGenerator(cycle([60, 61, 64, 66])))
        cf_2 = ChordField(
            quarter_duration=6,
            midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])),
            long_ending_mode='self_extend')

        cfg.add_child(cf_1)
        cfg.add_child(cf_2)

        xml_path = path + 'test_8.xml'
        simple_format = SimpleFormat()
        simple_format.extend(cf_1.simple_format)
        simple_format.extend(cf_2.simple_format)
        simple_format.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_9(self):
        # fields: duration_generators
        # group: midi_generator
        cfg = ChordField(
        )
        cf_1 = ChordField(
            quarter_duration=3,
            duration_generator=ValueGenerator(ArithmeticProgression(a1=0.5, an=1, correct_s=True))
        )
        cf_2 = ChordField(
            quarter_duration=6,
            duration_generator=ValueGenerator(ArithmeticProgression(a1=1, an=1, correct_s=True))
        )
        cfg.add_child(cf_1)
        cfg.add_child(cf_2)
        cfg.midi_generator = ValueGenerator(cycle([60]))
        simple_format = cfg.simple_format
        self.score.set_time_signatures(quarter_durations=ceil(cfg.quarter_duration))
        xml_path = path + 'test_9.xml'
        simple_format.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_10(self):
        midi_generator = ValueGenerator()

        midi_generator.add_child(
            ValueGenerator(RandomInterpolation(start=[60, 62, 66, 68], end=[67, 69, 73, 75], seed=10), duration=10)
        )

        midi_generator.add_child(
            ValueGenerator(RandomInterpolation(start=[67, 69, 73, 75], end=[60, 62, 66, 68], seed=11), duration=10)
        )

        cfg = ChordField(
        )
        cf_1 = ChordField(
            quarter_duration=3,
            duration_generator=ValueGenerator(ArithmeticProgression(a1=0.5, an=1, correct_s=True))
        )
        cf_2 = ChordField(
            quarter_duration=6,
            duration_generator=ValueGenerator(ArithmeticProgression(a1=1, an=1, correct_s=True))
        )

        cfg.add_child(cf_1)
        cfg.add_child(cf_2)
        cfg.midi_generator = midi_generator
        cfg.__next__()

    def test_11(self):
        cfg = ChordField(
            duration_generator=ValueGenerator(
                RandomInterpolation(start=[0.25, 0.25, 0.5], end=[0.5, 0.75, 1], seed=20)))
        cf_1 = ChordField(
            quarter_duration=3,
            midi_generator=ValueGenerator(cycle([60, 61, 64, 66])))
        cf_2 = ChordField(
            quarter_duration=6,
            midi_generator=ValueGenerator(cycle([72, 73, 74, 73, 72])),
            long_ending_mode='self_extend')

        cfg.add_child(cf_1)
        cfg.add_child(cf_2)
        xml_path = path + 'test_11.xml'
        copied = cfg.__deepcopy__()
        copied.simple_format.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_12(self):
        # breathing: duration_generator automatically, midi_generator: InterpolationGroup(2 x RandomInterpolations)
        breathing_proportions = [3, 7, 3, 10, 3]
        breathing_break_points = [1.5, 0.2, 1.5]
        breathing = Breathe(quarter_duration=sum(breathing_proportions), proportions=breathing_proportions,
                            breakpoints=breathing_break_points)

        midi_generator = ValueGenerator()

        midi_generator.add_child(
            ValueGenerator(RandomInterpolation(start=[60, 62, 66, 68], end=[67, 69, 73, 75], seed=10), duration=10)
        )

        midi_generator.add_child(
            ValueGenerator(RandomInterpolation(start=[67, 69, 73, 75], end=[60, 62, 66, 68], seed=11), duration=10)
        )

        breathing.midi_generator = midi_generator
        copied = breathing.__deepcopy__()
        xml_path = path + 'test_12.xml'
        self.score.set_time_signatures(quarter_durations=breathing_proportions)
        copied.simple_format.to_stream_voice().add_to_score(self.score)

        self.score.max_division = 5
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_13(self):
        proportions = (1, 3, 1, 5, 1)
        breakpoints = (1, Fraction(1, 7), 1)
        breathe = Breathe(proportions=proportions, breakpoints=breakpoints, quarter_duration=20)
        expected = 20
        actual = breathe.quarter_duration
        self.assertEqual(expected, actual)

    def test_14(self):
        proportions = (1, 3, 1, 5, 1)
        breakpoints = (1, Fraction(1, 7), 1)
        breathe = Breathe(proportions=proportions, breakpoints=breakpoints, quarter_duration=20, quantize=1)
        expected = [2, 5, 2, 9, 2]
        actual = [child.quarter_duration for child in breathe.children]
        self.assertEqual(expected, actual)

    def test_15(self):
        field = ChordField()
        child_1 = field.add_child(ChordField(
            midi_generator=ValueGenerator(cycle([71])),
            duration_generator=ValueGenerator(cycle([1])),
            quarter_duration=1))
        child_2 = field.add_child(ChordField(
            midi_generator=ValueGenerator(cycle([71])),
            duration_generator=ValueGenerator(cycle([1])),
            quarter_duration=1))
        actual = field.chords
        expected = child_1.chords + child_2.chords
        self.assertEqual(expected, actual)



import os

from musicscore.musicxml.elements.note import Notehead

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):

    def test_1(self):
        fm = FractalMusic(tempo=60, quarter_duration=10)
        fm.midi_generator.midi_range = [60, 72]
        fm.add_layer()
        for child in fm.get_children():
            child.add_gliss(grid=0.5, show_heads=True)
            child.chord_field.chords[0].add_articulation('accent')
        score = fm.get_score(show_fractal_orders=True, layer_number=fm.number_of_layers)
        xml_path = path + '_test_1.xml'
        score.write(path=xml_path)
        self.assertCompareFiles(xml_path)

    def test_2(self):
        fm = FractalMusic(tempo=60, quarter_duration=10, multi=(1, 2))
        fm.midi_generator.midi_range = [60, 72]
        fm.add_layer()
        for child in fm.get_children():
            child.chord.add_articulation('accent')
            child.chord.add_dynamics('f')
        for child in fm.get_children():
            child.add_gliss()
        score = fm.get_score(show_fractal_orders=True, layer_number=fm.number_of_layers)
        xml_path = path + '_test_2.xml'
        score.write(path=xml_path)
        self.assertCompareFiles(xml_path)

    def test_3(self):
        fm = FractalMusic(tempo=60, quarter_duration=10, multi=(1, 3))
        fm.midi_generator.midi_range = [60, 72]
        fm.add_layer()
        for child in fm.get_children():
            for midi in child.chord.midis:
                midi.notehead = Notehead('diamond', filled='no')
        for child in fm.get_children():
            child.add_gliss()

        fm.get_children()[-1].chord_field.chords[-1].get_post_grace_chords()[0].midis[0].notehead = Notehead('diamond',
                                                                                                             filled='no')
        score = fm.get_score(show_fractal_orders=True, layer_number=fm.number_of_layers)
        xml_path = path + '_test_3.xml'
        score.write(path=xml_path)
        self.assertCompareFiles(xml_path)



import itertools
import os
from math import ceil, floor

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.elements.note import Notations, Notehead
from musicscore.musicxml.types.complextypes.notations import Slide
from prettytable import PrettyTable
from quicktions import Fraction

from musurgia import basic_functions, scaledvalues
from musurgia.chordfield.chordfield import ChordField
from musurgia.chordfield.valuegenerator import ValueGenerator
from musurgia.fractaltree.fractaltree import FractalTree
from musurgia.fractaltree.midigenerators import RelativeMidi, MidiGenerator
from musurgia.interpolation import Interpolation
from musurgia.pdf.segmentedline import SegmentedLine
from musurgia.permutation import permute
from musurgia.quantize import get_quantized_values
from musurgia.timing import Timing


class FractalMusicException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class SetDurationFirstException(FractalMusicException):
    def __init__(self, *args):
        super().__init__('set duration first!', *args)


class SetTempoFirstException(FractalMusicException):
    def __init__(self, *args):
        super().__init__('set tempo first!', *args)


class TempoIsAlreadySet(FractalMusicException):
    def __init__(self, *args):
        super().__init__('FractalMusic().tempo can only be set once', *args)


class ChildTempoIsAlreadySet(FractalMusicException):
    def __init__(self, *args):
        super().__init__('FractalMusic().tempo of parent can not be set after setting tempo of fractal_tree', *args)


class MergeException(FractalMusicException):
    def __init__(self, *args):
        super().__init__(*args)


class MergeTempoException(MergeException):
    def __init__(self, tempo, *args):
        super().__init__('nodes to merge must have the same tempo {}'.format(tempo), *args)


class ChordFieldConflict(FractalMusicException):
    def __init__(self, *args):
        super().__init__('node.chord and node.chord_field cannot be both set!', *args)


class SimpleFormatConflict(FractalMusicException):
    def __init__(self, *args):
        super().__init__('simple_format and node.chord or node.chord_field cannot be both set!', *args)


class NoneChordException(FractalMusicException):
    def __init__(self, *args):
        super().__init__('node.chord and node.chord_field are both None.', *args)


class FractalMusic(FractalTree):
    def __init__(self,
                 duration=None,
                 midi_generator=None,
                 tempo=None,
                 quarter_duration=None,
                 tree_directions=None,
                 permute_directions=True,
                 *args, **kwargs):

        # super().__init__(value=duration, *args, **kwargs)
        super().__init__(*args, **kwargs)

        self._children_generated_midis = None
        self._chord = None
        self._chord_field = None
        self._midi_generator = None
        self._midi_value = None
        self._permute_directions = True
        self._simple_format = None
        self._tempo = None
        self._tree_directions = None

        self.midi_generator = midi_generator
        if duration:
            self.duration = duration

        self.tempo = tempo
        self.permute_directions = permute_directions
        self.tree_directions = tree_directions

        self.quarter_duration = quarter_duration

    # //private properties
    @property
    def _children_midis(self):
        return [child.midi_value for child in self.get_children()]

    @property
    def _midi_iterator(self):
        mg = self.midi_generator
        return mg.iterator

    @_midi_iterator.setter
    def _midi_iterator(self, value):
        raise AttributeError('_midi_iterator cannot be set directly. Use midi-generator!')

    # //private methods
    def _calculate_durational_position_in_tree(self):
        parent = self.up
        if self.is_root:
            return 0
        else:
            index = parent.get_children().index(self)
            if index == 0:
                return parent.durational_position_in_tree
            else:
                return parent.get_children()[index - 1].durational_position_in_tree + parent.get_children()[
                    index - 1].duration

    def _check_merge_nodes(self, nodes):
        tempo = set([node.tempo for node in nodes])
        if len(tempo) != 1:
            raise MergeTempoException(tempo)

    def _child_has_tempo(self):
        children_tempi = [child.tempo for child in self.get_children() if child.tempo]
        if children_tempi:
            return True
        else:
            return False

    # //public properties

    @property
    def auto_midi_range(self):
        if self.is_root:
            raise AttributeError('root has no auto_midi_range')

        parent_midis = self.up.children_generated_midis
        self_index = self.up.get_children().index(self)
        auto_range = parent_midis[self_index:self_index + 2]
        return auto_range

    @property
    def children_generated_midis(self):
        if self._children_generated_midis is None:
            self._children_generated_midis = []
            self._children_generated_midis = list(self._midi_iterator)

        return [midi for midi in self._children_generated_midis]

    @property
    def chord(self):
        if self.chord_field is None and self._simple_format is None:
            if self._chord is None:
                self._chord = TreeChord(quarter_duration=self.quarter_duration, midis=[self.midi_value])
            else:
                self._chord.quarter_duration = self.quarter_duration
            return self._chord
        else:
            return None

    @property
    def chord_field(self):
        if self._chord_field is not None:
            list(self._chord_field)
        return self._chord_field

    @chord_field.setter
    def chord_field(self, chord_field):
        if chord_field is not None and not isinstance(chord_field, ChordField):
            raise TypeError()
        if chord_field is not None:
            chord_field.quarter_duration = self.quarter_duration
        self._chord_field = chord_field

    @property
    def duration(self):
        return self.value

    @duration.setter
    def duration(self, val):
        self.value = val

    @property
    def durational_position_in_tree(self):
        return self._calculate_durational_position_in_tree()

    @property
    def midi_generator(self):
        if self._midi_generator is None:
            try:
                if not self.duration:
                    raise SetDurationFirstException()
                self._midi_generator = RelativeMidi(midi_range=[71, 71], proportions=self.children_fractal_values,
                                                    directions=None)
                self._midi_generator.node = self
            except AttributeError:
                raise Exception('midi_generator is None and cannot be set:')
        else:
            self._midi_generator.node = self

        if isinstance(self._midi_generator, RelativeMidi):

            if not self._midi_generator._directions:
                tree_directions = self.get_root().tree_directions

                if self.permute_directions:
                    directions = permute(tree_directions, self.permutation_order)
                else:
                    directions = tree_directions

                self._midi_generator.directions = directions

            if self._midi_generator.midi_range is None:
                if self.is_root:
                    self._midi_generator.midi_range = [71]
                else:
                    self._midi_generator.midi_range = self.auto_midi_range
                    self._midi_generator._auto_ranged = True

        return self._midi_generator

    @midi_generator.setter
    def midi_generator(self, value):
        if value is not None:
            if not (isinstance(value, MidiGenerator)):
                err = 'midi_generator can only be an instance of subclasses of MidiGenerator or None. None=RelativeMidi(midi_range=self.midi_range, proportions=permute(self.fractal_proportions, self.fractal.permutation_order)'
                raise TypeError(err)

        self._midi_generator = value
        if self._midi_generator:
            self._midi_generator.node = self

    @property
    def midi_value(self):
        if self._midi_value is None:
            if self.is_root:
                self._midi_value = 71
            else:
                self._midi_value = self.up.children_generated_midis[self.up.get_children().index(self)]
        return self._midi_value

    @midi_value.setter
    def midi_value(self, value):
        if value is not None:
            if isinstance(value, int) or isinstance(value, float):
                if value >= 18 or value == 0:
                    self._midi_value = value
                else:
                    raise ValueError('midi can be 0 and greater than 18')
            else:
                raise TypeError('midi can only be int, float or None not {}'.format(type(value)))
        else:
            self._midi_value = value

    @property
    def multi(self):
        return super().multi

    @multi.setter
    def multi(self, val):
        FractalTree.multi.fset(self, val)
        # is midi_generator already set?
        try:
            if self._midi_generator:
                if isinstance(self._midi_generator, RelativeMidi):
                    midi_range = self._midi_generator.midi_range
                    _auto_ranged = self._midi_generator._auto_ranged
                    self._midi_generator = self._midi_generator.copy()
                    self._midi_generator.midi_range = midi_range
                    self._midi_generator._auto_ranged = _auto_ranged
                else:
                    self._midi_generator = self._midi_generator.copy()
        except AttributeError:
            pass

    @property
    def permute_directions(self):
        return self._permute_directions

    @permute_directions.setter
    def permute_directions(self, value):
        if not isinstance(value, bool):
            raise TypeError()
        self._permute_directions = value

    @property
    def quarter_duration(self):
        if not self.tempo:
            raise SetTempoFirstException()
        return Fraction(Fraction(self.duration * self.tempo), Fraction(60))

    @quarter_duration.setter
    def quarter_duration(self, val):
        if val is not None:
            if not self.tempo:
                raise SetTempoFirstException()
            self.duration = Fraction(Fraction(val * 60), Fraction(self.tempo))

    @property
    def quarter_position_in_tree(self):
        return self.position_in_tree * self.tempo / 60

    @property
    def simple_format(self):
        raise FractalMusicException('simple_format is not gettable.')

    @simple_format.setter
    def simple_format(self, val):
        if not isinstance(val, SimpleFormat):
            raise TypeError()
        self._simple_format = val

    @property
    def tempo(self):
        return self._tempo

    @tempo.setter
    def tempo(self, val):
        if self._tempo:
            raise TempoIsAlreadySet()
        if self._child_has_tempo():
            raise ChildTempoIsAlreadySet()
        self._tempo = val
        for child in self.get_children():
            child.tempo = val

    @property
    def tree_directions(self):
        return self._tree_directions

    @tree_directions.setter
    def tree_directions(self, val):
        if val is None:
            val = [1, -1]
        length = len(self.permutation_order)
        if len(val) > length:
            val = val[:length]
        elif len(val) < length:
            output = []
            cycled = itertools.cycle(val)
            for i in range(length):
                output.append(cycled.__next__())
            val = output
        self._tree_directions = val
        for node in self.traverse():
            if node._midi_generator:
                try:
                    if self.permute_directions and node._midi_generator.directions:
                        node._midi_generator.directions = permute(self.tree_directions, self.permutation_order)
                    else:
                        node._midi_generator.directions = self.tree_directions
                except AttributeError:
                    pass

    # //public methods
    # add
    def add_gliss(self, unit=1, grid=1, min_quarter_duration=Fraction(1, 8), show_heads=False):
        def get_slides(chord):
            return [slide for notation in chord.get_children_by_type(Notations) for slide in
                    notation.get_children_by_type(Slide)]

        position = self.quarter_position_in_tree
        delta = unit - (position - int(position))
        if delta >= min_quarter_duration:
            duration_generator = ValueGenerator(itertools.chain(iter([delta]), itertools.cycle([unit])))
        else:
            duration_generator = ValueGenerator(itertools.cycle([unit]))

        chord_field = ChordField(duration_generator=duration_generator,
                                 midi_generator=ValueGenerator(
                                     Interpolation(start=self.midi_generator.midi_range[0],
                                                   end=self.midi_generator.midi_range[1], grid=grid)),
                                 long_ending_mode='cut',
                                 short_ending_mode='stretch')

        first_chord = self.chord.__deepcopy__()
        self.chord_field = chord_field
        list(chord_field)
        first_chord.quarter_duration = chord_field.chords[0].quarter_duration
        chord_field.chords[0] = first_chord
        if chord_field.chords[-1].quarter_duration < min_quarter_duration:
            chord_field._chords = chord_field._chords[:-1]

        for chord in chord_field.chords[1:]:
            for midi in chord.midis:
                if show_heads is False:
                    midi.notehead = Notehead('none')
                    alter = midi.get_pitch_rest().alter
                    if alter:
                        midi.transpose(-alter.value)
                    midi.accidental.force_hide = True

        if self.previous_leaf and self.previous_leaf.chord_field:
            list(self.previous_leaf.chord_field)
            previous_slides_types = [slide.type for slide in get_slides(self.previous_leaf.chord_field.chords[0])]
            if 'start' in previous_slides_types:
                try:
                    self.chord.add_slide('stop')
                except AttributeError:
                    self.chord_field.chords[0].add_slide('stop')
        if self.next_leaf is None:
            grace = TreeChord(midis=self.midi_generator.midi_range[1])
            grace.add_slide('stop')
            chord_field.chords[-1].add_grace_chords([grace], mode='post')

        chord_field.chords[0].add_slide('start')

    def add_info(self, *show_attributes, condition=None):
        def write_words(words, placement, relative_y):
            if placement is None:
                placement = 'below'
            if relative_y is None:
                if placement == 'below':
                    relative_y = -15
                else:
                    relative_y = 15
            try:
                self.chord.add_words(words, placement=placement, relative_y=relative_y)
            except AttributeError:
                try:
                    for chord in self.chord_field.chords:
                        chord.add_words(words, placement=placement, relative_y=relative_y)
                except AttributeError:
                    for chord in self._simple_format.chords:
                        chord.add_words(words, placement=placement, relative_y=relative_y)

        def get_words(attr):
            if attr == 'fractal_order':
                text = [node.fractal_order for node in self.get_branch()[1:]]
                words = ('\n'.join([str(x) for x in text]))
            elif isinstance(attr, str):
                words = str(getattr(self, attr))
            elif callable(attr):
                words = str(attr(self))
            else:
                raise AttributeError()
            return words

        for attr in show_attributes:
            placement = None
            relative_y = None
            if isinstance(attr, tuple):
                words = get_words(attr[0])
                placement = attr[1]
                try:
                    relative_y = attr[2]
                except IndexError:
                    pass
            else:
                words = get_words(attr)

            write_words(words, placement, relative_y)

        for child in self.get_children():
            if condition is None or condition(child) is True:
                child.add_info(*show_attributes, condition=condition)

    # get

    def get_choral(self, range_factor=1, direction=None, last=False):
        choral_midis = list(dict.fromkeys(
            self.get_choral_midis(range_factor=range_factor, direction=direction, last=last)
        ))
        return TreeChord(quarter_duration=self.quarter_duration, midis=choral_midis)

    def get_choral_midis(self, range_factor=1, direction=None, last=False):
        range_ = (self.midi_generator.midi_range[1] - self.midi_generator.midi_range[0]) * range_factor
        self_direction = range_ / abs(range_)
        if direction is None or direction == 0:
            pass
        elif direction == 1:
            range_ = abs(range_)
        elif direction == -1:
            range_ = -abs(range_)

        children_midis = self.children_generated_midis
        choral_midis = []

        step = 2 / self.midi_generator.microtone
        if self_direction > 0:
            scale = scaledvalues.ScaledValues(min(children_midis), max(children_midis), self.midi_value,
                                              self.midi_value + range_,
                                              step=step)
        else:
            scale = scaledvalues.ScaledValues(max(children_midis), min(children_midis), self.midi_value,
                                              self.midi_value + range_,
                                              step=step)
        if not last:
            min_midi = min(children_midis)
            max_midi = max(children_midis)
            if self.midi_value == min_midi:
                children_midis.remove(max_midi)
            else:
                children_midis.remove(min_midi)
        for midi in children_midis:
            new_midi = scale(midi)
            choral_midis.append(new_midi)

        return choral_midis

    def get_children_score(self, score=None, show_fractal_orders=False, show_midis=False):
        if not score:
            score = self.get_score_template()

        old_tempo = None
        max_layers = max([child.number_of_layers for child in self.get_children()])
        for child in self.get_children():
            tempo = child.tempo
            if tempo == old_tempo:
                show_metronome = False
            else:
                show_metronome = True
            old_tempo = tempo
            child_score = child.get_score(layer_number=list(range(0, max_layers + 1)),
                                          show_fractal_orders=show_fractal_orders, show_midis=show_midis,
                                          barline='light-light', show_metronome=show_metronome)

            score.extend(child_score)

        score.accidental_mode = 'modern'
        score.get_measure(-1).set_barline_style('light-heavy')
        return score

    def get_root_score(self, score=None, layer_number=None, show_fractal_orders=False, show_midis=False
                       # , show_positions=False
                       ):

        if not score:
            score = self.get_score_template()

        try:
            return self.get_score(score=score, layer_number=layer_number, show_fractal_orders=show_fractal_orders,
                                  show_midis=show_midis)
        except SetTempoFirstException as err:
            if not self.get_children():
                raise err
            else:
                for child in self.get_children():
                    score.extend(
                        child.get_score(layer_number=layer_number, show_fractal_orders=show_fractal_orders,
                                        show_midis=show_midis, barline='light-light'))

        # if set_time_signatures:
        #     if layer_number == 0 or not self.get_children():
        #         score.set_time_signatures(quarter_durations=self.quarter_duration, barline_style='light-light', times=times)
        #     else:
        #         durations = [fractal_tree.quarter_duration for fractal_tree in self.get_children() if fractal_tree.quarter_duration != 0]
        #         score.set_time_signatures(
        #             durations=durations, barline_style='light-light', times=times)
        # else:

        score.accidental_mode = 'modern'

        # if show_positions:
        #     for node in self.get_children()[1:]:
        #         position_in_tree = float(node.position_in_tree)
        #         quarter_position_in_tree = position_in_tree * self.tempo / 60
        #         position_in_score = quarter_position_in_tree * 60 / self.score_tempo
        #         node.chord.add_words(
        #             Timing.get_clock(time=round(position_in_score, 1), mode='msreduced'),
        #             enclosure='rectangle', relative_y=40)
        score.get_measure(-1).set_barline_style('light-heavy')
        return score

    def get_score_template(self):
        score = TreeScoreTimewise()
        score.tuplet_line_width = 2.4
        score.page_style.orientation = 'landscape'
        score.page_style.system_distance = 180
        score.page_style.staff_distance = 150
        score.page_style.top_system_distance = 150
        score.page_style.bottom_margin = 100

        score.add_title('module: {}'.format(self.name))
        rounded_duration = round(float(self.duration), 1)
        if rounded_duration == int(rounded_duration):
            rounded_duration = int(rounded_duration)

        clock = Timing.get_clock(rounded_duration, mode='msreduced')
        score.add_subtitle(
            'duration: {}'.format(clock))
        score.accidental_mode = 'modern'
        return score

    def get_score(self, score=None, layer_number=None, show_fractal_orders=False, show_midis=None,
                  barline='light-heavy',
                  show_metronome=True):
        if not score:
            score = self.get_score_template()
        score.set_time_signatures(quarter_durations=self.quarter_duration)
        if show_fractal_orders:
            for node in self.traverse():
                if node.fractal_order is not None:
                    try:
                        node.chord.add_lyric(node.fractal_order)
                    except AttributeError:
                        try:
                            for chord in node.chord_field.chords:
                                chord.add_lyric(node.fractal_order)
                        except AttributeError:
                            for chord in node._simple_format.chords:
                                chord.add_lyric(node.fractal_order)
        if show_midis:
            for node in self.traverse():
                try:
                    node.chord.add_words(node.chord.midis)
                except AttributeError:
                    try:
                        for chord in node.chord_field.chords:
                            chord.add_words(chord.midis)
                    except AttributeError:
                        for chord in node.simple_format.chords:
                            chord.add_words(chord.midis)

        def layer_to_score(layer_number, part_number):
            try:
                sf = self.get_simple_format(layer_number)
                sf.auto_clef()
                v = sf.to_stream_voice(1)
                v.add_to_score(score, part_number=part_number)
            except ValueError:
                print('module {}: number_of_layers={}: getting layer {} not possible'.format(self.name,
                                                                                             self.number_of_layers,
                                                                                             i))

        if layer_number is None:
            for i, j in zip(range(self.number_of_layers + 1), range(1, self.number_of_layers + 2)):
                layer_to_score(i, j)

        elif hasattr(layer_number, '__iter__'):
            for i, j in zip(layer_number, range(1, len(layer_number) + 1)):
                layer_to_score(i, j)
        else:
            layer_to_score(layer_number, 1)

        if show_metronome:
            score.get_measure(1).get_part(1).add_metronome(per_minute=self.tempo, relative_y=40)
        score.get_measure(-1).set_barline_style(barline)
        return score

    def get_simple_format(self, layer=None):
        if layer is None:
            layer = self.get_farthest_leaf().get_distance()
        self.get_layer(layer=layer)

        simple_format = SimpleFormat()
        for node in basic_functions.flatten(self.get_layer(layer)):
            if node.chord is not None:
                if node.chord_field is not None:
                    raise ChordFieldConflict()
                if node._simple_format is not None:
                    raise SimpleFormatConflict()
                copied_chord = node.chord.__deepcopy__()
                simple_format.add_chord(copied_chord)
            elif node.chord_field is not None:
                if node._simple_format is not None:
                    raise SimpleFormatConflict()
                for chord in node.chord_field.chords:
                    copied_chord = chord.__deepcopy__()
                    simple_format.add_chord(copied_chord)
            elif node._simple_format is not None:
                for chord in node._simple_format.chords:
                    copied_chord = chord.__deepcopy__()
                    simple_format.add_chord(copied_chord)

            else:
                raise NoneChordException()

        return simple_format

    # set
    def set_chord(self, chord):
        self._chord = chord

    def set_none_tempi(self, val=60):
        try:
            self.tempo = val
        except TempoIsAlreadySet:
            pass
        except ChildTempoIsAlreadySet:
            for child in self.get_children():
                child.set_none_tempi(val=val)

    # others
    def change_midis(self):
        if not isinstance(self._midi_generator, RelativeMidi):
            raise TypeError(
                'set_reduced_auto_ranges can only be applied to FractalMusic nodes with RelativeMidi as midi_generator')
        else:

            directions = self.midi_generator.directions
            midi_range = self.midi_generator.midi_range
            microtone = self.midi_generator.microtone
            # print('inside'+ midi_range)
            # children = iter(self.get_children())
            # old_generated_midis = self.children_generated_midis
            # print("old_generated_midis", old_generated_midis)
            # self._children_generated_midis = [node.midi_value for node in self.get_children()]
            # self._children_generated_midis.append(old_generated_midis[-1])
            # print("new_generated_midis", old_generated_midis)

            self.midi_generator = None
            self.midi_generator.proportions = None
            self.midi_generator.midi_range = midi_range
            self.midi_generator.microtone = microtone
            self.midi_generator.directions = directions
            self._children_generated_midis = None

            for child in self.get_children():
                child.midi_value = None
                child._chord = None

            for child in self.get_children():
                if isinstance(child.midi_generator, RelativeMidi):
                    child.midi_generator.midi_range = None
                    child.midi_generator.directions = None
                    child.midi_generator.proportions = None
                    child.midi_generator._iterator = None
                    child._children_generated_midis = None

    def change_quarter_duration(self, new_quarter_duration):
        if new_quarter_duration is not None:
            if not self.tempo:
                raise SetTempoFirstException()
            try:
                self.change_value(Fraction(new_quarter_duration * 60, self.tempo))
            except TypeError:
                self.change_value(Fraction(Fraction(new_quarter_duration * 60), Fraction(self.tempo)))

    def create_segmented_line_group(self, layers=None, *show_attributes):
        # show_attributes: attr / (attr) / (attr, function/kwdict) / (attr, function, kwdict)
        def add_line_text_labels():
            for line, ch in zip(segmented_line.line_segments, self.get_children()):
                for attr_tuple in show_attributes:
                    attr_func = None
                    kwdict = None
                    if isinstance(attr_tuple, tuple):
                        for x in attr_tuple[1:]:
                            if isinstance(x, str):
                                attr_func = lambda y: getattr(ch, x)
                            if callable(x):
                                attr_func = x
                            elif isinstance(x, dict):
                                kwdict = x
                    else:
                        attr_func = lambda child: getattr(child, x)

                    label = line.start_mark_line.add_text_label(str(attr_func(ch)))
                    if kwdict:
                        for key in kwdict.keys():
                            setattr(label, key, kwdict[key])

        segmented_line_group = SegementedLineGroup()
        segmented_line_group.add_segemented_line(SegmentedLine(lengths=[ch.duration for ch in self.get_children()]))
        for segmented_line in segmented_line_group.get_segmented_lines():
            add_line_text_labels()
        return segmented_line_group

    def find_best_tempo_in_range(self, min_tempo=40, max_tempo=144):
        min_quarter_duration = ceil(Timing(duration=self.duration, tempo=min_tempo).quarter_duration)
        max_quarter_duration = floor(Timing(duration=self.duration, tempo=max_tempo).quarter_duration)
        tempi = [Timing(duration=self.duration, quarter_duration=x).tempo for x in
                 range(min_quarter_duration, max_quarter_duration + 1)]
        tempo_deltas = [abs(round(tempo) - tempo) for tempo in tempi]
        min_delta = min(tempo_deltas)
        index = tempo_deltas.index(min_delta)
        best_tempo = int(round(tempi[index]))
        return best_tempo

    def find_best_tempi_in_list(self, tempi):
        quarter_durations = [Timing(duration=self.duration, tempo=tempo).quarter_duration for tempo in tempi]
        quarter_duration_deltas = [abs(round(quarter_duration) - quarter_duration) for quarter_duration in
                                   quarter_durations]
        rounded_duration_deltas = [round(float(delta), 2) for delta in quarter_duration_deltas]

        min_delta = min(rounded_duration_deltas)
        indices = [i for i, quarter_duration_delta in enumerate(rounded_duration_deltas) if
                   quarter_duration_delta == min_delta]
        best_tempi = [tempi[index] for index in indices]
        return best_tempi

    def inverse_tree_directions(self):
        self.tree_directions = [direction * -1 for direction in self.tree_directions]

    def split(self, *proportions):
        if hasattr(proportions[0], '__iter__'):
            proportions = proportions[0]

        proportions = [Fraction(prop) for prop in proportions]

        for prop in proportions:
            duration = self.quarter_duration * prop / sum(proportions)
            new_node = self.copy()
            new_node.multi = self.multi
            new_node.quarter_duration = duration
            new_node._fractal_order = self.fractal_order
            new_node.midi_value = self.midi_value
            new_node.midi_generator.midi_range = self.midi_generator.midi_range
            if self._chord:
                new_node.set_chord(self.chord.__deepcopy__())
            self.add_child(new_node)

        return self.get_children()

    def quantize_leaves(self, grid_size):

        #     quantizing quarter_durations!
        leaves = list(self.traverse_leaves())
        quarter_durations = [leaf.quarter_duration for leaf in leaves]
        quantized_quarter_durations = get_quantized_values(quarter_durations, grid_size)
        for leaf, quantized_duration in zip(leaves, quantized_quarter_durations):
            leaf.change_quarter_duration(quantized_duration)

    def quantize_children(self, grid_size=0.5):
        #     quantizing quarter_durations!
        children = self.get_children()
        quarter_durations = [child.quarter_duration for child in children]
        quantized_quarter_durations = get_quantized_values(quarter_durations, grid_size)
        for child, quantized_duration in zip(children, quantized_quarter_durations):
            child.change_quarter_duration(quantized_duration)

    def recreate_segmented_line(self):
        self._segmented_line = None

    def reduce_children(self, condition):
        super().reduce_children(condition)

        try:
            self.change_midis()
        except TypeError:
            pass

    def round_leaves(self):
        #     quantizing quarter_durations!
        leaves = list(self.traverse_leaves())
        rounded_quarter_durations = [round(leaf.quarter_duration) for leaf in leaves]
        for leaf, rounded_quarter_duration in zip(leaves, rounded_quarter_durations):
            leaf.change_quarter_duration(rounded_quarter_duration)

    def round_children(self):
        children = self.get_children()
        rounded_quarter_durations = [round(child.quarter_duration) for child in children]
        for child, rounded_quarter_duration in zip(children, rounded_quarter_durations):
            child.change_quarter_duration(rounded_quarter_duration)

    def write_infos(self, file_name):
        os.system('touch ' + file_name)
        file = open(file_name, 'w')
        x = PrettyTable()

        x.field_names = ["name", "f_o", "quarters", "duration", "midi", "perm_ord", "multi", "directions",
                         "midi_range", "childr_midis"]

        for node in self.traverse():
            x.add_row([node.name, node.fractal_order, round(float(node.quarter_duration), 2),
                       Timing(quarter_duration=float(node.quarter_duration), tempo=self.tempo).clock,
                       node.midi_value,
                       node.permutation_order, node.multi, node.midi_generator.directions,
                       node.midi_generator.midi_range, node.children_generated_midis]
                      )

        if self.is_root:
            file.write('name: root')
        else:
            file.write('name: {} fo: {}'.format(self.name, self.fractal_order))
        file.write('\n')
        file.write(
            'quarter_duration: {}, tempo: {}, duration: {}'.format(round(float(self.quarter_duration), 2),
                                                                   self.tempo,
                                                                   round(float(self.duration), 2)))
        file.write('\n')
        file.write('tree directions: {}'.format(self.tree_directions))
        file.write('\n')
        file.write('\n')
        file.write(x.get_string())
        file.close()

    # //copy
    def copy(self):
        copied_midi_generator = self.midi_generator.copy()

        copied = super().copy()
        copied.tempo = self.tempo
        copied.midi_generator = copied_midi_generator
        copied.tree_directions = self.tree_directions
        copied.permute_directions = self.permute_directions

        if self.fertile is False:
            copied._midi_value = self._midi_value
        return copied

    def __deepcopy__(self, memodict={}):
        copied = super().__deepcopy__()
        copy_attribute_names = ['_chord_field', '_chord', '_midi_value', '_simple_format', '_midi_generator',
                                '_children_generated_midis', '_tree_directions', '_permute_directions', '_tempo']

        self.deepcopy_attributes(copied, copy_attribute_names)

        return copied

    def __copy__(self):
        copied = self.__class__(proportions=self.proportions, tree_permutation_order=self.tree_permutation_order)
        copied.duration = self.duration
        copied.midi_generator = None
        copied._tempo = self.tempo
        copied.quarter_duration = self.quarter_duration
        copied.tree_directions = self.tree_directions
        copied.permute_directions = self.permute_directions
        return copied



from pathlib import Path

import matplotlib as mpl
from matplotlib.afm import AFM


class FontError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def _make_afm_path_dictionary():
    def check_entry():
        old_afm = output.get((family, weight, style))
        if old_afm is not None:
            old_header = old_afm._header
            new_header = afm._header
            diff = set(old_header) ^ set(new_header)
            if diff == {b'CapHeight'}:
                if new_header.get(b'CapHeight'):
                    return True
                #
                # print('old:', old_header.get(b'CapHeight'))
                # print('new:', new_header.get(b'CapHeight'))
            elif diff == set():
                return False
            else:
                raise AttributeError(
                    f'{family},  {weight}, {style} already in dict: {old_afm} differnce: {diff}')
        else:
            return True

    output = {}
    directory = Path(mpl.get_data_path(), 'fonts', 'afm')
    for file in directory.iterdir():
        afm_path = file
        with afm_path.open('rb') as fh:
            afm = AFM(fh)
        family = afm.get_familyname()
        weight = afm.get_weight().lower()
        if afm.get_angle() < 0:
            style = 'italic'
        else:
            style = 'regular'
        if check_entry():
            output[family, weight, style] = afm

    return output


class Font:
    __AFM_PATH_DICTIONARY = _make_afm_path_dictionary()

    _FAMILY = ['Helvetica', 'Courier', 'Times']
    _WEIGHT = ['bold', 'medium']
    _STYLE = ['italic', 'regular']

    def __init__(self, family='Helvetica', weight='medium', style='regular', size=10, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._family = None
        self._weight = None
        self._style = None
        self._size = None
        self._afm = None

        self.family = family
        self.weight = weight
        self.style = style
        self.size = size

    def _set_afm(self):
        if self.family and self.weight and self.style:
            self._afm = self.__AFM_PATH_DICTIONARY[self.family, self.weight, self.style]

    @property
    def family(self):
        return self._family

    @family.setter
    def family(self, val):
        if val not in self._FAMILY:
            raise FontError(f'{val} not a valid value: current valid families are: {self._FAMILY}')
        self._family = val
        self._set_afm()

    @property
    def weight(self):
        return self._weight

    @weight.setter
    def weight(self, val):
        if val not in self._WEIGHT:
            raise FontError('{} not a valid value: {}'.format(val, self._WEIGHT))
        self._weight = val
        self._set_afm()

    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, val):
        if val not in self._STYLE:
            raise FontError('{} not a valid value: {}'.format(val, self._STYLE))
        self._style = val
        self._set_afm()

    @property
    def size(self):
        return self._size

    @size.setter
    def size(self, val):
        self._size = val

    def get_text_pixel_width(self, val):
        return (self._afm.string_width_height(val)[0] / 1000) * self.size

    def get_text_pixel_height(self, val):
        return (self._afm.string_width_height(val)[1] / 1000) * self.size



class ScaledValues(object):
    def __init__(self, old_min, old_max, new_min, new_max, step=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.old_min = old_min
        self.old_max = old_max
        self.new_min = new_min
        self.new_max = new_max
        self._step = None
        self.step = step

    @property
    def step(self):
        return self._step

    @step.setter
    def step(self, val):
        if val is not None and val <= 0:
            raise TypeError('step.value must positive not {}'.format(type(val)))
        self._step = val

    def __call__(self, x):
        y = self.new_min + (self.new_max - self.new_min) * (x - self.old_min) / (self.old_max - self.old_min)
        if self.step:
            y = round(y / self.step) * self.step
        return y



from quicktions import Fraction
from unittest import TestCase

from musurgia.basic_functions import xToD, dToX
from musurgia.quantize import get_quantized_values


class Test(TestCase):

    def test_1(self):
        input = [0.2, 0.333, 0.6, 0.99, 0.1, 0.5]
        example = get_quantized_values(input, 0.6)
        result = [Fraction(0, 1), Fraction(3, 5), Fraction(3, 5), Fraction(6, 5), Fraction(0, 1), Fraction(0, 1)]
        self.assertEqual(result, example)

    def test_2(self):
        midis = [60, 61.5, 58.5, 56.5, 63]
        intervals = xToD(midis)
        quantized_intervals = get_quantized_values(intervals, 1)
        quantized_midis = [int(x) for x in dToX(quantized_intervals, first_element=midis[0])]
        result = [60, 61, 58, 57, 63]
        self.assertEqual(result, quantized_midis)



from quicktions import Fraction

from musurgia.pdf.labeled import Labeled
from musurgia.pdf.masterslave import Master, Slave
from musurgia.pdf.newdrawobject import DrawObject


class StraightLine(Slave, DrawObject):
    def __init__(self, mode, length, show=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mode = None
        self._length = None

        self.mode = mode
        self.length = length
        self.show = show

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, val):
        permitted = ['h', 'horizontal', 'v', 'vertical']
        if val not in permitted:
            raise ValueError(f'mode.value {val} must be in {permitted}')
        self._mode = val

    @property
    def length(self):
        return self._length

    @length.setter
    def length(self, val):
        if not isinstance(val, float) and not isinstance(val, int) and not isinstance(val, Fraction):
            raise TypeError(f"length.value must be of type float, int or Fraction  not{type(val)}")
        self._length = val

    @staticmethod
    def get_opposite_mode(mode):
        if mode == 'h':
            return 'v'
        elif mode == 'v':
            return 'h'
        elif mode == 'horizontal':
            return 'vertical'
        elif mode == 'vertical':
            return 'horizontal'
        else:
            raise AttributeError()

    def get_relative_x2(self):
        if self.mode in ['h', 'horizontal']:
            return self.relative_x + self.length
        else:
            return self.relative_x

    def get_relative_y2(self):
        if self.mode in ['v', 'vertical']:
            return self.relative_y + self.length
        else:
            return self.relative_y

    def draw(self, pdf):
        if self.show:
            with pdf.prepare_draw_object(self):
                x2 = self.get_relative_x2() - self.relative_x
                y2 = self.get_relative_y2() - self.relative_y
                pdf.line(0, 0, x2, y2)


class MarkLine(StraightLine, Labeled):
    def __init__(self, placement, length=3, *args, **kwargs):
        super().__init__(length=length, *args, **kwargs)
        self._placement = None
        self.placement = placement

    @property
    def placement(self):
        return self._placement

    @placement.setter
    def placement(self, val):
        permitted = ['start', 'end']
        if val not in permitted:
            raise ValueError(f'placement.value {val} must be in {permitted}')
        self._placement = val

    def get_middle_y(self):
        return self.length / 2

    def draw(self, pdf):
        with pdf.saved_state():
            self.draw_above_text_labels(pdf)
            self.draw_left_text_labels(pdf)
            super().draw(pdf)
            if self.below_text_labels:
                pdf.translate(0, 2)
                self.draw_below_text_labels(pdf)


class LineSegment(Master, DrawObject):
    def __init__(self, mode, length, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._straight_line = StraightLine(name='straight_line', mode=mode, length=length, master=self)
        marker_mode = StraightLine.get_opposite_mode(self.mode)
        self._start_mark_line = MarkLine(name='start_mark_line', mode=marker_mode, master=self, placement='start')
        self._end_mark_line = MarkLine(name='end_mark_line', mode=marker_mode, master=self, placement='end', show=False)

    @property
    def straight_line(self):
        return self._straight_line

    @property
    def start_mark_line(self):
        return self._start_mark_line

    @property
    def end_mark_line(self):
        return self._end_mark_line

    @property
    def mode(self):
        return self.straight_line.mode

    @property
    def length(self):
        return self.straight_line.length

    def _get_straight_line_margin(self, margin):
        if margin in ['l', 'left']:
            return 0
        elif margin in ['r', 'right']:
            return 0
        elif margin in ['t', 'top']:
            return 0
        elif margin in ['b', 'bottom']:
            return 0
        else:
            raise AttributeError(margin)

    def _get_mark_line_margin(self, margin):
        if margin in ['l', 'left']:
            return 0
        elif margin in ['r', 'right']:
            return 0
        elif margin in ['t', 'top']:
            return 0
        elif margin in ['b', 'bottom']:
            return 0
        else:
            raise AttributeError(margin)

    def _get_straight_line_position(self, position):
        if position not in ['x', 'y']:
            raise AttributeError(position)

        if self.mode in ['h', 'horizontal']:
            if position == 'x':
                return 0
            elif position == 'y':
                return 0
                # return max([ml.get_middle_y() for ml in [self.start_mark_line, self.end_mark_line]])
        else:
            if position == 'x':
                return 0
                # return max([ml.get_middle_y() for ml in [self.start_mark_line, self.end_mark_line]])
            elif position == 'y':
                return 0

    def _get_mark_line_position(self, position, mark_line):
        if position not in ['x', 'y']:
            raise AttributeError(position)

        if mark_line.mode in ['h', 'horizontal']:
            if position == 'x':
                # return 0
                return -mark_line.length / 2
            else:
                if mark_line.placement == 'start':
                    return 0
                else:
                    return self.length
        else:
            if position == 'y':
                # return 0
                return -mark_line.length / 2
            else:
                if mark_line.placement == 'start':
                    return 0
                else:
                    return self.length

    def get_slave_margin(self, slave, margin):
        if slave.name == 'straight_line':
            return self._get_straight_line_margin(margin)
        elif slave.name == 'start_mark_line':
            return self._get_mark_line_margin(margin)
        elif slave.name == 'end_mark_line':
            return self._get_mark_line_margin(margin)
        else:
            raise AttributeError(slave)

    def get_slave_position(self, slave, position):
        if slave.name == 'straight_line':
            return self._get_straight_line_position(position)
        elif slave.name == 'start_mark_line':
            return self._get_mark_line_position(position, slave)
        elif slave.name == 'end_mark_line':
            return self._get_mark_line_position(position, slave)
        else:
            raise AttributeError(slave)


class HorizontalLineSegment(LineSegment):
    def __init__(self, length, *args, **kwargs):
        super().__init__(mode='horizontal', length=length, *args, **kwargs)

    def get_relative_x2(self):
        return self.relative_x + self.length

    def get_relative_y2(self):
        return self.relative_y + max([ml.get_height() for ml in [self.start_mark_line, self.end_mark_line]])

    def draw(self, pdf):
        with pdf.prepare_draw_object(self):
            self.start_mark_line.draw(pdf)
            self.straight_line.draw(pdf)
            self.end_mark_line.draw(pdf)


class HorizontalSegmentedLine(DrawObject):
    def __init__(self, lengths, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._segments = None
        self._make_segments(lengths)

    @property
    def lengths(self):
        return [segment.length for segment in self.segments]

    @property
    def segments(self):
        return self._segments

    def _make_segments(self, lengths):
        if not lengths:
            raise AttributeError('lengths must be set.')
        self._segments = [HorizontalLineSegment(length) for length in lengths]
        self._segments[-1].end_mark_line.show = True

    def get_relative_x2(self):
        return self.relative_x + sum(self.lengths)

    def get_relative_y2(self):
        return self.relative_y + max([segment.get_height() for segment in self.segments])

    def draw(self, pdf):
        with pdf.prepare_draw_object(self):
            for segment in self.segments:
                segment.draw(pdf)
                pdf.translate(segment.get_width(), 0)


class VerticalLineSegment(LineSegment):
    def __init__(self, length, *args, **kwargs):
        super().__init__(mode='vertical', length=length, *args, **kwargs)

    def get_relative_x2(self):
        return self.relative_x

    def get_relative_y2(self):
        return self.relative_y + self.length

    def draw(self, pdf):
        with pdf.prepare_draw_object(self):
            self.start_mark_line.draw(pdf)
            self.straight_line.draw(pdf)
            self.end_mark_line.draw(pdf)


class VerticalSegmentedLine(DrawObject):
    def __init__(self, lengths, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._segments = None
        self._make_segments(lengths)

    @property
    def lengths(self):
        return [segment.length for segment in self.segments]

    @property
    def segments(self):
        return self._segments

    def _make_segments(self, lengths):
        if not lengths:
            raise AttributeError('lengths must be set.')
        self._segments = [VerticalLineSegment(length) for length in lengths]
        self._segments[-1].end_mark_line.show = True

    def get_relative_x2(self):
        return self.relative_x + max([segment.get_height() for segment in self.segments])

    def get_relative_y2(self):
        return self.relative_y + sum(self.lengths)

    def draw(self, pdf):
        with pdf.prepare_draw_object(self):
            for segment in self.segments:
                segment.draw(pdf)
                pdf.translate(0, segment.get_height())



from pathlib import Path

from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase, create_test_path

path = Path(__file__)


class Test(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()

    def test_line(self):
        pdf_path = create_test_path(path, 'line.pdf')
        self.pdf.rect(0, 0, 50, 50)
        self.pdf.clip_rect(0, 0, 50, 50)
        self.pdf.line(10, 20, 100, 100)
        self.pdf.write(pdf_path)
        self.assertCompareFiles(pdf_path)

    def test_line_break(self):
        pdf = Pdf(unit='pt')

        def draw_scene():
            pdf.line(0, 0, 300, 50)
            pdf.rect(0, 0, 20, 30)
            pdf.rect(50, 20, 2, 3)
            pdf.rect(152, 28, 4, 6)
            pdf.line(300, 50, 600, 0)

        pdf_path = create_test_path(path, 'line_break.pdf')
        pdf.translate(20, -20)

        pdf.rect(0, 0, 150, 50)
        draw_scene()

        pdf.translate(0, -100)
        pdf.rect(0, 0, 150, 50)
        with pdf.saved_state():
            pdf.clip_rect(0, 0, 150, 50)
            pdf.translate(-150, 0)
            pdf.set_draw_color(255, 0, 0)
            draw_scene()

        pdf.translate(0, -100)
        pdf.rect(0, 0, 150, 50)
        with pdf.saved_state():
            pdf.clip_rect(0, 0, 150, 50)
            pdf.translate(-300, 0)
            pdf.set_draw_color(0, 0, 255)
            draw_scene()

        pdf.write(pdf_path)
        self.assertCompareFiles(pdf_path)



class Random(object):
    import random
    current_random = random

    def __init__(self, pool=None, periodicity=None, forbidden_list=None, seed=None):
        self._pool = None
        self._periodicity = None
        self._forbidden_list = None
        self._seed = None
        self._counter = 0
        self._result = None

        self.pool = pool
        self.periodicity = periodicity
        self.forbidden_list = forbidden_list
        self.seed = seed

    # periodicity None will sets the periodicity always to len(pool)-2
    @property
    def pool(self):
        return self._pool

    @pool.setter
    def pool(self, values):
        if values is not None:
            try:
                self._pool = list(dict.fromkeys(values))
            except TypeError:
                self._pool = [values]

    @property
    def periodicity(self):
        return self._periodicity

    @periodicity.setter
    def periodicity(self, value):
        if value is not None:
            if isinstance(value, int) and value >= 0:
                self._periodicity = value
            else:
                raise ValueError()

    @property
    def forbidden_list(self):
        if not self._forbidden_list:
            self._forbidden_list = []
        return self._forbidden_list

    @forbidden_list.setter
    def forbidden_list(self, values):
        self._forbidden_list = values

    @property
    def seed(self):
        return self._seed

    @seed.setter
    def seed(self, value):
        self._seed = value
        self.current_random.seed(value)

    @property
    def counter(self):
        return self._counter

    @property
    def result(self):
        if self._result is None:
            self._result = []
        return self._result

    @property
    def iterator(self):

        if not self.pool:
            raise AttributeError('set pool')

        while True:
            periodicity = self.periodicity
            if self.periodicity is None:
                periodicity = len(self.pool) - 2

            elif self.periodicity >= len(self.pool):
                periodicity = len(self.pool) - 1

            if periodicity < 0: periodicity = 0

            def check(x):

                def forbid_element(x):
                    if len(self.forbidden_list) >= periodicity:
                        self.forbidden_list.pop(0)
                    self.forbidden_list.append(x)

                if periodicity != 0:
                    if x in self.forbidden_list:
                        return False
                    else:
                        forbid_element(x)
                        return True
                else:
                    return True

            if len(self.forbidden_list) > periodicity:
                # print "self.periodicity", self.periodicity
                self.forbidden_list = self.forbidden_list[(-1 * periodicity):]

            random_element = self.pool[self.current_random.randrange(len(self.pool))]
            while check(random_element) is False:
                random_element = self.pool[self.current_random.randrange(len(self.pool))]

            yield random_element

    def __next__(self):
        next_el = self.iterator.__next__()
        self._counter += 1
        # self._sum+=next_el
        self.result.append(next_el)

        return next_el

    def __deepcopy__(self, memodict={}):
        copied = self.__class__(pool=self.pool, periodicity=self.periodicity, forbidden_list=self.forbidden_list,
                                seed=self.seed)
        return copied



import os

from musurgia.table.table import Table
from musurgia.unittest import TestCase

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def test_1(self):
        main_table = Table(hrules=1)
        main_table.field_names = [1, 2, 3]
        main_table.add_row(['a', 'b', 'c'])
        side_table = Table()
        side_table.add_row(['aa', 'bb', 'cc'])
        main_table.extend(side_table)
        table_path = path + '_test_1.txt'

        main_table.writ_to_path(table_path=table_path, title='test extend')
        self.assertCompareFiles(table_path)



import os
from itertools import cycle

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from quicktions import Fraction

from musurgia.unittest import TestCase
from musurgia.chordfield.chordfield import Breathe, ChordField
from musurgia.chordfield.valuegenerator import ValueGenerator
from musurgia.fractaltree.fractalmusic import FractalMusic
from musurgia.interpolation import RandomInterpolation

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()
        self.fm = FractalMusic(quarter_duration=24, tempo=40)

    def test_1(self):
        proportions = (1, 3, 1, 5, 1)
        breakpoints = (1, Fraction(1, 7), 1)
        quarter_durations = [8, 12]
        breathe = Breathe(proportions=proportions, breakpoints=breakpoints, quarter_duration=sum(quarter_durations),
                          quantize=1)
        breathe.midi_generator = ValueGenerator(cycle([71]))
        test_chord_field = ChordField(duration_generator=breathe.duration_generator.__deepcopy__())
        for i in range(len(quarter_durations)):
            quarter_duration = quarter_durations[i]
            midi = 60 + i
            test_chord_field.add_child(
                ChordField(midi_generator=ValueGenerator(cycle([midi])), long_ending_mode='self_extend',
                           short_ending_mode='self_shrink', quarter_duration=quarter_duration))

        test_chord_field_2 = ChordField(duration_generator=breathe.duration_generator.__deepcopy__())
        for i in range(len(quarter_durations)):
            quarter_duration = quarter_durations[i]
            midi = 72 + i
            test_chord_field_2.add_child(
                ChordField(midi_generator=ValueGenerator(cycle([midi])), long_ending_mode='cut',
                           short_ending_mode='add_rest', quarter_duration=quarter_duration))
        breathe.simple_format.to_stream_voice().add_to_score(score=self.score, part_number=1)
        test_chord_field.simple_format.to_stream_voice().add_to_score(score=self.score, part_number=2)

        simple_format = SimpleFormat()
        for child in test_chord_field_2.children:
            simple_format.extend(child.simple_format)

        simple_format.to_stream_voice().add_to_score(score=self.score, part_number=3)
        xml_path = path + '_test_1.xml'
        self.score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_2(self):
        self.fm.add_layer()
        self.fm.quantize_children(grid_size=1)
        proportions = (1, 3, 1, 5, 1)
        breakpoints = (1, Fraction(1, 7), 1)
        selected_nodes = self.fm.get_children()[:2]
        breath_quarter_duration = sum([node.quarter_duration for node in selected_nodes])
        breathe = Breathe(proportions=proportions, breakpoints=breakpoints, quarter_duration=breath_quarter_duration,
                          quantize=1)
        parent_chord_field = ChordField(duration_generator=breathe.duration_generator.__deepcopy__())
        for i in range(len(selected_nodes)):
            midi = 60 + i
            chord_field = ChordField(midi_generator=ValueGenerator(cycle([midi])), long_ending_mode='self_extend',
                                     short_ending_mode='self_shrink')
            parent_chord_field.add_child(chord_field)
            node = selected_nodes[i]
            node.chord_field = chord_field

        score = self.fm.get_score(show_fractal_orders=True, layer_number=self.fm.number_of_layers)
        xml_path = path + '_test_2.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_3(self):
        def get_children_midis(node):
            copy = node.__deepcopy__()
            copy.midi_generator.midi_range = node.midi_generator.midi_range
            copy.add_layer()

            output = [child.chord.midis[0].value for child in copy.get_children()]
            return output

        self.fm.multi = (self.fm.multi[0], self.fm.multi[1] + 1)
        self.fm.midi_generator.midi_range = (60, 72)
        self.fm.add_layer()
        self.fm.quantize_children(grid_size=1)
        proportions = (1, 3, 1, 3, 1)
        breakpoints = (1, Fraction(1, 7), 1)
        selected_nodes = self.fm.get_children()[:2]
        breath_quarter_duration = sum([node.quarter_duration for node in selected_nodes])
        breathe = Breathe(proportions=proportions, breakpoints=breakpoints, quarter_duration=breath_quarter_duration,
                          quantize=1)
        parent_chord_field = ChordField(duration_generator=breathe.duration_generator.__deepcopy__())
        for i in range(len(selected_nodes)):
            node = selected_nodes[i]
            start_chord = get_children_midis(node)
            next_node = node.next_sibling
            if next_node:
                end_chord = get_children_midis(next_node)
            else:
                end_chord = start_chord

            chord_field = ChordField(
                midi_generator=ValueGenerator(RandomInterpolation(start=start_chord, end=end_chord, seed=10)),
                long_ending_mode='self_extend',
                short_ending_mode='self_shrink')
            parent_chord_field.add_child(chord_field)

            node.chord_field = chord_field

        score = self.fm.get_score(show_fractal_orders=True, layer_number=self.fm.number_of_layers)
        score.max_division = 7
        xml_path = path + '_test_3.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_4(self):
        # self.fm.multi = (self.fm.multi[0], self.fm.multi[1] + 1)
        def get_chord_stamp():
            node = fm.__copy__()
            node.midi_generator.midi_range = fm.midi_generator.midi_range
            node.midi_generator.microtone = 4
            node.add_layer()
            chord_midis = [child.midi_value for child in node.get_children()]
            output = [chord_midi - chord_midis[node.size // 2] for chord_midi in chord_midis]
            return output

        fm = FractalMusic(proportions=(1, 2, 3, 4, 5, 6, 7), tree_permutation_order=(2, 6, 4, 1, 3, 7, 5),
                          quarter_duration=30, tempo=80)
        fm.midi_generator.midi_range = (60, 72)
        fm.midi_generator.microtone = 4
        fm.add_layer()
        fm.quantize_children(grid_size=1)
        proportions = (1, 10, 1, 7, 1)
        breakpoints = (1, Fraction(1, 7), 1)
        selected_nodes = fm.get_children()[1:3]

        breath_quarter_duration = sum([node.quarter_duration for node in selected_nodes])
        breathe = Breathe(proportions=proportions, breakpoints=breakpoints, quarter_duration=breath_quarter_duration,
                          quantize=1)
        parent_chord_field = ChordField(duration_generator=breathe.duration_generator.__deepcopy__())
        for i in range(len(selected_nodes)):
            node = selected_nodes[i]
            start_chord = [stamp + node.midi_value for stamp in get_chord_stamp()]
            next_node = node.next_sibling
            if next_node:
                end_chord = [stamp + node.midi_value for stamp in get_chord_stamp()]
            else:
                end_chord = start_chord

            chord_field = ChordField(
                midi_generator=ValueGenerator(RandomInterpolation(start=start_chord, end=end_chord, seed=10)),
                long_ending_mode='self_extend',
                short_ending_mode='self_shrink')
            parent_chord_field.add_child(chord_field)

            node.chord_field = chord_field

        score = fm.get_score(show_fractal_orders=True, layer_number=fm.number_of_layers)
        score.max_division = 7
        score.finish()
        partwise = score.to_partwise()
        xml_path = path + '_test_4.xml'
        partwise.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_5(self):

        def get_chord(node):
            output = fm.get_choral_midis(range_factor=1)
            transposition = node.midi_value - output[3]
            output = [midi + transposition for midi in output]
            # output = node.get_choral_midis(range_factor=8 - node.fractal_order)
            return output

        fm = FractalMusic(proportions=(1, 2, 3, 4, 5, 6, 7), tree_permutation_order=(2, 6, 4, 1, 3, 7, 5),
                          quarter_duration=30, tempo=72)
        # fm.multi = (self.fm.multi[0], self.fm.multi[1] + 1)
        fm.midi_generator.midi_range = (62, 62 + 11)
        fm.midi_generator.microtone = 4
        fm.add_layer()
        fm.quantize_children(grid_size=1)
        selected_nodes = fm.get_children()[2:5]

        proportions = (1, 10, 1, 7, 1)
        breakpoints = (1, Fraction(1, 7), 1)
        breath_quarter_duration = sum([node.quarter_duration for node in selected_nodes])
        breathe = Breathe(proportions=proportions, breakpoints=breakpoints, quarter_duration=breath_quarter_duration,
                          quantize=1)
        parent_chord_field = ChordField(duration_generator=breathe.duration_generator.__deepcopy__())
        for i in range(len(selected_nodes)):
            node = selected_nodes[i]
            start_chord = get_chord(node)
            next_node = node.next_sibling
            if next_node:
                end_chord = get_chord(next_node)
            else:
                end_chord = start_chord

            chord_field = ChordField(
                midi_generator=ValueGenerator(RandomInterpolation(start=start_chord, end=end_chord, seed=10)),
                long_ending_mode='self_extend',
                short_ending_mode='self_shrink')
            parent_chord_field.add_child(chord_field)

            node.chord_field = chord_field

        score = fm.get_score(show_fractal_orders=True, layer_number=fm.number_of_layers)
        score.max_division = 7
        score.finish()
        partwise = score.to_partwise()
        xml_path = path + '_test_5.xml'
        partwise.write(xml_path)
        self.assertCompareFiles(xml_path)

    def test_6(self):
        def get_chord(node):
            output = fm.get_choral_midis(range_factor=1)
            transposition = node.midi_value - output[3]
            output = [midi + transposition for midi in output]
            # output = node.get_choral_midis(range_factor=8 - node.fractal_order)
            return output

        fm = FractalMusic(proportions=(1, 2, 3, 4, 5, 6, 7), tree_permutation_order=(2, 6, 4, 1, 3, 7, 5),
                          quarter_duration=30, tempo=72)
        # fm.multi = (self.fm.multi[0], self.fm.multi[1] + 1)
        fm.midi_generator.midi_range = (62, 62 + 11)
        fm.midi_generator.microtone = 4
        fm.add_layer()
        fm.quantize_children(grid_size=1)

        def make_breathe(nodes, proportions, breakpoints):
            breath_quarter_duration = sum([node.quarter_duration for node in nodes])
            breathe = Breathe(proportions=proportions, breakpoints=breakpoints,
                              quarter_duration=breath_quarter_duration,
                              quantize=1)

            parent_chord_field = ChordField(duration_generator=breathe.duration_generator.__deepcopy__())
            for i in range(len(nodes)):
                node = selected_nodes[i]
                start_chord = get_chord(node)
                next_node = node.next_sibling
                if next_node:
                    end_chord = get_chord(next_node)
                else:
                    end_chord = start_chord

                chord_field = ChordField(
                    quarter_duration=node.quarter_duration,
                    midi_generator=ValueGenerator(RandomInterpolation(start=start_chord, end=end_chord, seed=10)),
                    long_ending_mode='self_extend',
                    short_ending_mode='self_shrink')
                parent_chord_field.add_child(chord_field)

            return parent_chord_field

        selected_nodes = fm.get_children()[2:5]
        # print(sum([node.quarter_duration for node in selected_nodes]))
        proportions = (1, 10, 1, 7, 1)
        breakpoints = (1, Fraction(1, 7), 1)
        breathe = make_breathe(nodes=selected_nodes, proportions=proportions,
                               breakpoints=breakpoints)
        # print(breathe.quarter_duration)
        fm.merge_children(2, 3, 2)
        # print(fm.get_children()[1].quarter_duration)
        fm.get_children()[1].simple_format = breathe.simple_format
        score = fm.get_score(show_fractal_orders=True, layer_number=fm.number_of_layers)
        score.max_division = 7
        score.finish()
        partwise = score.to_partwise()
        xml_path = path + '_test_6.xml'
        partwise.write(xml_path)
        self.assertCompareFiles(xml_path)



import os
from unittest import TestCase

from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.fractaltree.fractalmusic import FractalMusic

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        fm = FractalMusic(proportions=[1, 2, 3], tree_permutation_order=[3, 1, 2], duration=12, tempo=72)
        fm.quarter_duration = round(fm.quarter_duration)
        fm.add_layer()
        quarter_durations = [float(leaf.quarter_duration) for leaf in fm.traverse_leaves()]
        # print(quarter_durations)
        # print(sum(quarter_durations))
        fm.quantize_leaves(grid_size=0.5)
        quarter_durations = [float(leaf.quarter_duration) for leaf in fm.traverse_leaves()]
        self.assertEqual(quarter_durations, [7.0, 2.5, 4.5])
        # print([leaf.chord.quarter_duration for leaf in fm.traverse_leaves()])



import os

from musurgia.table.table import Table
from musurgia.fractaltree.fractalmusic import FractalMusic
from musurgia.fractaltree.fractaltree import FractalTreeException
from prettytable import PrettyTable
from quicktions import Fraction


class Module(FractalMusic):
    def __init__(self, row_number=None, column_number=None, *args, **kargs):
        super().__init__(*args, **kargs)
        self.row_number = row_number
        self.column_number = column_number
        self._parent_row = None
        self._parent_column = None
        self._name = None

    @property
    def parent_square(self):
        return self.parent_row.parent_square

    @property
    def parent_row(self):
        return self._parent_row

    @property
    def parent_column(self):
        return self._parent_column

    def set_name(self, value):
        self._name = value

    @property
    def name(self):
        if self._name:
            return self._name

        if self.row_number and self.column_number:
            return str(self.row_number) + '_' + str(self.column_number)
        else:
            return super().name

    def write_info(self, text_path, show_quarter_durations=False):
        os.system('touch ' + text_path)
        file = open(text_path, 'w')
        file.write("module_tempo: " + str(self.tempo))
        file.write("\n")
        x = PrettyTable(hrules=1)

        leaf_names = [leaf.name for leaf in self.traverse_leaves()]
        leaf_fractal_orders = [leaf.fractal_order for leaf in self.traverse_leaves()]

        leaf_durations = [leaf.duration for leaf in self.traverse_leaves()]

        rounded_leaf_durations = [round(float(dur), 2) for dur in leaf_durations]

        leaf_quarter_durations = [leaf.quarter_duration for leaf in self.traverse_leaves()]
        rounded_quarter_durations = [round(float(dur), 2) for dur in leaf_quarter_durations]

        x.field_names = ["name", "info", *leaf_names, "sum"]
        x.add_row([self.name, 'frac_ord', *leaf_fractal_orders, " "])
        x.add_row([self.name, 'durations', *rounded_leaf_durations, round(float(sum(leaf_durations)), 2)])
        if show_quarter_durations:
            x.add_row(
                [self.name, 'quarter_dur', *rounded_quarter_durations,
                 round(float(sum(leaf_quarter_durations)), 2)])

        file.write(x.get_string())
        file.close()

    def __deepcopy__(self, memodict={}):
        copied = super().__deepcopy__(memodict)
        copied.row_number = self.row_number
        copied.column_number = self.column_number
        copied._name = self._name

        return copied


class RowColumn(object):
    def __init__(self, parent_square, number, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parent_square = None
        self.parent_square = parent_square
        self._number = None
        self.number = number
        self._modules = []

    @property
    def parent_square(self):
        return self._parent_square

    @parent_square.setter
    def parent_square(self, value):
        if not isinstance(value, Square):
            raise TypeError('parent_square.value must be of type Square not {}'.format(type(value)))
        self._parent_square = value

    @property
    def number(self):
        return self._number

    @number.setter
    def number(self, value):
        if not isinstance(value, int):
            raise TypeError('number.value must be of type int not {}'.format(type(value)))
        if not 0 < value <= self.parent_square.side_size:
            raise ValueError('number.value must be of between 0 and {}'.format(self.parent_square.side_size))
        self._number = value

    @property
    def modules(self):
        return self._modules

    def get_module(self, number):
        return self.modules[number - 1]

    @property
    def quarter_duration(self):
        return sum([module.quarter_duration for module in self.modules])


class Row(RowColumn):
    def __init__(self, square, number, *args, **kwargs):
        super().__init__(square, number, *args, **kwargs)
        self._name = str(self.number)

    def _add_module(self, module):
        module._parent_row = self
        self._modules.append(module)

    def set_tempo(self, tempo):
        for module in self.modules:
            module.tempo = tempo

    def set_name(self, val):
        self._name = val

    @property
    def name(self):
        return self._name

    def change_module_quarter_duration(self, module_number, new_quarter_duration):
        factor = Fraction(Fraction(new_quarter_duration),
                          Fraction(self.modules[module_number - 1].quarter_duration))

        for index, module in enumerate(self.modules):
            module.duration = self.modules[index].duration * factor

    def change_module_duration(self, module_number, new_duration):
        factor = Fraction(
            Fraction(new_duration),
            Fraction(self.modules[module_number - 1].duration)
        )

        for index, module in enumerate(self.modules):
            module.duration = self.modules[index].duration * factor


class Column(RowColumn):
    def __init__(self, square, number, *args, **kwargs):
        super().__init__(square, number, *args, **kwargs)

    def _add_module(self, module):
        module._parent_column = self
        self._modules.append(module)


class Square(object):
    def __init__(self, duration, proportions, tree_permutation_order, first_multi=(1, 1),
                 reading_direction='horizontal', name=None):

        self._duration = None
        self._proportions = None
        self._modules = {}
        self._tree_permutation_order = None
        self._first_multi = (1, 1)
        self._reading_direction = None
        self._side_size = None
        self._name = None
        self._parent_square_group = None

        self.duration = duration
        self.proportions = proportions
        self.reading_direction = reading_direction
        self.tree_permutation_order = tree_permutation_order
        self.first_multi = first_multi
        self._rows = None
        self._columns = None
        self.name = name

    # //private methods
    def _calculate_module_values(self):
        if self.duration is not None and self.proportions is not None and self.tree_permutation_order is not None:
            row_durations = [self.duration * prop / float(sum(self.proportions)) for prop in self.proportions]
            for (row, column) in [(i + 1, j + 1) for i in range(self.side_size) for j in range(self.side_size)]:
                module_durations = [row_durations[row - 1] * prop / float(sum(self.proportions)) for prop in
                                    self.proportions]
                multi = self.index_to_r_c(
                    self.r_c_to_index(row, column) + self.r_c_to_index(self.first_multi[0], self.first_multi[1]))
                module = Module(duration=module_durations[column - 1],
                                tree_permutation_order=self.tree_permutation_order, proportions=self.proportions,
                                multi=multi, reading_direction=self.reading_direction)
                (module.row_number, module.column_number) = (row, column)
                self._modules[(row, column)] = module
                module._parent_square = self

    # //public properties
    @property
    def columns(self):
        if self._columns is None:
            self._columns = []
            for column_number in range(1, self.side_size + 1):
                column = Column(square=self, number=column_number)
                for row_number in range(1, self.side_size + 1):
                    module = self.get_module(row_number, column_number)
                    column._add_module(module)

                self._columns.append(column)
        return self._columns

    @property
    def duration(self):
        if self.modules != {}:
            durations = map(lambda module: module.duration, self.modules.values())
            self._duration = sum(durations)
        return self._duration

    @duration.setter
    def duration(self, value):
        if value is None:
            raise ValueError('duration cannot be None')
        self._duration = value
        self._calculate_module_values()

    @property
    def first_multi(self):
        return self._first_multi

    @first_multi.setter
    def first_multi(self, value):
        self._first_multi = value
        if self._modules != {}:
            for key in self._modules.keys():
                module = self._modules[key]
                module.multi = self.index_to_r_c(
                    self.r_c_to_index(module.row_number, module.column_number) + self.r_c_to_index(self.first_multi[0],
                                                                                                   self.first_multi[1]))

    @property
    def index_in_group(self):
        if self.parent_square_group:
            return self.parent_square_group.squares.index(self)
        else:
            return None

    @property
    def modules(self):
        return self._modules

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        self._name = val

    @property
    def parent_square_group(self):
        return self._parent_square_group

    @parent_square_group.setter
    def parent_square_group(self, val):
        if not isinstance(val, SquareGroup):
            raise TypeError('parent_square_group.value must be of type SquareGroup  not{}'.format(type(val)))
        self._parent_square_group = val

    @property
    def proportions(self):
        return self._proportions

    @proportions.setter
    def proportions(self, values):
        if values is None:
            raise ValueError('proportions cannot be None')
        if self.side_size is not None:
            if len(values) != self.side_size:
                raise ValueError('wrong proportions length')
        else:
            self._side_size = len(values)
        self._proportions = values
        self._calculate_module_values()

    @property
    def reading_direction(self):
        return self._reading_direction

    @reading_direction.setter
    def reading_direction(self, val):
        if self._reading_direction:
            raise FractalTreeException('reading_direction can only be set during initialisation')
        permitted = ['horizontal', 'vertical']
        if val not in permitted:
            raise ValueError('reading_direction.value {} must be in {}'.format(val, permitted))
        self._reading_direction = val

    @property
    def rows(self):
        if self._rows is None:
            self._rows = []
            for row_number in range(1, self.side_size + 1):
                row = Row(square=self, number=row_number)
                for column_number in range(1, self.side_size + 1):
                    module = self.get_module(row_number, column_number)
                    row._add_module(module)

                self._rows.append(row)
        return self._rows

    @property
    def side_size(self):
        return self._side_size

    @property
    def tree_permutation_order(self):
        return self._tree_permutation_order

    @tree_permutation_order.setter
    def tree_permutation_order(self, values):
        if values is None:
            raise ValueError('tree_permutation_order cannot be None')
        if self.side_size is not None:
            if len(values) != self.side_size:
                raise ValueError('wrong tree_permutation_order length')
        else:
            self._side_size = len(values)
        self._tree_permutation_order = values
        self._calculate_module_values()

    # //public methods

    # get
    def get_all_modules(self):
        return self.modules.values()

    def get_column(self, column_number):
        return self.columns[column_number - 1]

    def get_module(self, *args):
        args = tuple(args)
        return self.modules[args]

    def get_row(self, row_number):
        return self.rows[row_number - 1]

    # other
    def change_module_quarter_duration(self, row_number, column_number, new_quarter_duration):
        factor = Fraction(Fraction(new_quarter_duration),
                          Fraction(self.get_module(row_number, column_number).quarter_duration))

        for key in self.modules:
            self.modules[key].duration = self.modules[key].duration * factor

    def change_module_duration(self, row_number, column_number, new_duration):
        factor = Fraction(
            Fraction(new_duration),
            Fraction(self.get_module(row_number, column_number).duration)
        )

        for key in self.modules:
            self.modules[key].duration = self.modules[key].duration * factor

    def index_to_r_c(self, index):
        row = int(index / self.side_size) % self.side_size + 1
        column = index % self.side_size + 1
        return row, column

    def round_quarter_durations(self):
        for module in self.modules.values():
            module.quarter_duration = round(module.quarter_duration)

    def r_c_to_index(self, row, column):
        index = ((row - 1) * self.side_size) + (column - 1)
        return index

    def write_to_table(self, table=None, show_attributes=None):
        if not show_attributes:
            show_attributes = []

        if not table:
            table = Table(hrules=1)
            column_numbers = [str(number) for number in range(1, self.side_size + 1)]
            table.field_names = ['row', "column:", *column_numbers]
        elif not isinstance(table, Table):
            raise TypeError('table must be of type Table not {}'.format(type(table)))

        for row_number in range(1, self.side_size + 1):
            row = self.get_row(row_number)

            def _round(duration):
                output = round(duration, 1)
                if output != int(output):
                    output = float(output)
                return output

            row_name = row.name
            for attr in show_attributes:
                if attr == 'quarter_duration':
                    tempi = [module.tempo for module in row.modules]
                    if None in tempi:
                        raise Exception('set tempi first')
                    quarter_durations = [round(float(module.quarter_duration), 2) for module in row.modules]
                    table.add_row([row_name, 'quarter_dur', *quarter_durations])
                elif attr == 'duration':
                    durations = [_round(module.duration) for module in row.modules]
                    table.add_row([row_name, 'duration', *durations])
                else:
                    if isinstance(attr, str):
                        module_attributes = [getattr(module, attr) for module in row.modules]
                        table.add_row([row_name, attr, *module_attributes])
                    elif isinstance(attr, tuple) and len(attr) == 2 and isinstance(attr[0], str) and callable(
                            attr[1]):
                        module_attributes = [attr[1](module) for module in row.modules]
                        table.add_row([row_name, attr[0], *module_attributes])
                    else:
                        raise ValueError('attribute {} can not be written to table'.format(attr))
                row_name = ''
        return table

    def write_info(self, text_path, show_attributes=None, title=None):
        if not show_attributes:
            show_attributes = []

        os.system('touch ' + text_path)
        file = open(text_path, 'w')
        if title:
            file.write(title + '\n')
        table = self.write_to_table(show_attributes=show_attributes)

        file.write(table.get_string())
        file.close()

    # //copy
    def __deepcopy__(self, memodict={}):
        copied = self.__class__(duration=self.duration, proportions=self.proportions,
                                tree_permutation_order=self.tree_permutation_order, first_multi=self.first_multi,
                                reading_direction=self.reading_direction, name=self.name)

        for key in self.modules.keys():
            copied.modules[key] = self.modules[key].__deepcopy__()

        for row, copied_row in zip(self.rows, copied.rows):
            if row.name:
                copied_row.set_name(row.name)

        return copied


class SquareGroup(object):
    def __init__(self, *squares, name=None):
        super().__init__()
        self._name = None
        self._squares = []
        for square in squares:
            self.add_square(square)

        self.name = name

    @property
    def squares(self):
        return self._squares

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        self._name = val

    def add_square(self, val):
        if not isinstance(val, Square):
            raise TypeError()
        self._squares.append(val)
        val.parent_square_group = self

    def get_all_rows(self):
        return [row for square in self.squares for row in square.rows]

    def write_to_table(self, table=None, show_attributes=None):
        first_square = self.squares[0]
        if not table:
            table = first_square.write_to_table(table=None, show_attributes=show_attributes)
        for square in self.squares[1:]:
            table.extend(square.write_to_table(table=None, show_attributes=show_attributes))
        return table

    def write_info(self, text_path, show_attributes=None, title=None):

        os.system('touch ' + text_path)
        file = open(text_path, 'w')
        if title:
            file.write(title + '\n')
        table = self.write_to_table(show_attributes=show_attributes)

        file.write(table.get_string())
        file.close()

    def __deepcopy__(self, memodict={}):
        copied = self.__class__(*[square.__deepcopy__() for square in self.squares])
        return copied



from musurgia.pdf.text import TextLabel


class NameLabel(TextLabel):
    def __init__(self, text, relative_x=-10, relative_y=0, *args, **kwargs):
        super().__init__(text=text, relative_x=relative_x, relative_y=relative_y, *args, **kwargs)


class Named(object):
    def __init__(self, name=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._name = None
        self.name = name

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        if val is not None and not isinstance(val, NameLabel):
            val = NameLabel(val)
        self._name = val
        if self.name:
            self.name.parent = self
            self.name.relative_y += self.relative_y

    def __deepcopy__(self, memodict={}):
        return self.__class__(name=self.name.__deepcopy__())



import os

from musicscore.musictree.treescoretimewise import TreeScoreTimewise

from musurgia.unittest import TestCase
from musurgia.fractaltree.fractalmusic import FractalMusic

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):

    def setUp(self) -> None:
        self.fm = FractalMusic(tempo=60, quarter_duration=12, tree_permutation_order=(3, 1, 4, 2),
                               proportions=[1, 2, 3, 4],
                               multi=(1, 1))
        self.fm.midi_generator.midi_range = [55, 55 + 24]
        self.fm.midi_generator.microtone = 4

    def add_infos(self, fm):
        for leaf in fm.traverse_leaves():
            leaf.chord.add_lyric(leaf.fractal_order, relative_y=20)
            leaf.chord.add_words(round(float(leaf.quarter_duration), 2), relative_y=30)
            leaf.chord.add_words(leaf.chord.midis[0].value, relative_y=50)

    def test_1(self):

        self.fm.midi_generator.set_directions(1, 1, 1, 1)
        self.fm.add_layer()

        self.add_infos(self.fm)
        self.fm.add_layer()
        self.add_infos(self.fm)
        score = TreeScoreTimewise()
        score.page_style.staff_distance = 150

        for layer in range(self.fm.number_of_layers + 1):
            sf = self.fm.get_simple_format(layer=layer)
            v = sf.to_stream_voice(1)
            v.add_to_score(score, layer + 1)

        xml_path = path + '_test_1.xml'
        score.write(path=xml_path)
        self.assertCompareFiles(actual_file_path=xml_path)

    # def test_2(self):
    #     self.fm.midi_generator.directions = [1]
    #     self.fm.add_layer()
    #     print([round(float(chord.quarter_duration), 2) for chord in self.fm.simple_format.chords])
    #     print(basic_functions.xToD(self.fm.children_generated_midis))
    #     print(basic_functions.xToD([chord.midis[0].value for chord in self.fm.simple_format.chords]))
    #
    # def test_3(self):
    #     self.fm.midi_generator.directions = [1]
    #     self.fm.add_layer()
    #     self.fm.add_layer()
    #
    #     for node in self.fm.get_children():
    #         print([round(float(chord.quarter_duration), 2) for chord in node.simple_format.chords])
    #         print(basic_functions.xToD(node.children_generated_midis))
    #         print(basic_functions.xToD([chord.midis[0].value for chord in node.simple_format.chords]))
    #
    # def test_4(self):
    #     def print_infos(node):
    #         print(node.fractal_order)
    #         print('directions', node.midi_generator.directions)
    #         print('durations', [round(float(v), 2) for v in node.children_fractal_values])
    #         print('midis', node.children_generated_midis)
    #         print('intervals', basic_functions.xToD(node.children_generated_midis))
    #
    #     self.fm.midi_generator.directions = [1]
    #     self.fm.add_layer()
    #
    #     print_infos(self.fm)
    #
    #     for leaf in self.fm.traverse_leaves():
    #         print_infos(leaf)
    #
    #     self.fm.add_layer()
    #     for leaf in self.fm.traverse_leaves():
    #         print_infos(leaf)

    # def test_5(self):
    #     def print_infos(node):
    #         print(node.fractal_order)
    #         print('directions', node.midi_generator.directions)
    #         print('durations', [round(float(v), 2) for v in node.children_fractal_values])
    #         print('midis', node.children_generated_midis)
    #         print('intervals', basic_functions.xToD(node.children_generated_midis))
    #
    #     self.fm.midi_generator.directions = [1, -1]
    #     self.fm.add_layer()
    #
    #     print_infos(self.fm)
    #
    #     for leaf in self.fm.traverse_leaves():
    #         print_infos(leaf)
    #
    #     self.fm.add_layer()
    #     for leaf in self.fm.traverse_leaves():
    #         print_infos(leaf)
    #
    # def test_6(self):
    #     permutation_order = [8, 11, 7, 12, 10, 13, 9, 4, 1, 3, 6, 2, 5]
    #     fm = FractalMusic(tree_permutation_order=permutation_order, duration=900, proportions=list(range(1, 14)))
    #     fm.midi_generator.midi_range = [48, 84]
    #     fm.midi_generator.directions = [1, 1, -1, -1]
    #
    #     def print_infos(node):
    #         print(node.fractal_order)
    #         # print('directions', node.midi_generator.directions)
    #         durations = [round(float(v), 2) for v in node.children_fractal_values]
    #         # print('durations', durations)
    #         # print('midis', node.children_generated_midis)
    #         intervals = [round(float(v), 2) for v in basic_functions.xToD(node.children_generated_midis)]
    #         # print('intervals', intervals)
    #         print('durations,intervals',
    #               [(duration, interval) for duration, interval in zip(durations, intervals)])
    #         # print('durations/intervals',
    #         #       [Fraction(duration, interval) for duration, interval in zip(durations, intervals)])
    #
    #     fm.add_layer()
    #
    #     # print_infos(fm)
    #
    #     for leaf in fm.traverse_leaves():
    #         print_infos(leaf)
    #
    #     # self.fm.add_layer()
    #     # for leaf in self.fm.traverse_leaves():
    #     #     print_infos(leaf)

    def test_7(self):
        fm = FractalMusic(tempo=60, proportions=[1, 2, 3, 4], tree_permutation_order=[3, 1, 4, 2], quarter_duration=20)
        fm.midi_generator.midi_range = [60, 79]
        fm.add_layer()
        fm.chord.add_words(fm.midi_generator.midi_range)

        for child in fm.get_children():
            child.chord.add_lyric(child.fractal_order)
            child.chord.add_words(child.midi_generator.midi_range)

        for index, node in enumerate(fm.get_children()):
            node.midi_generator.midi_range = [60 - index, 72 - index]

        fm.add_layer()
        for leaf in fm.traverse_leaves():
            leaf.chord.add_lyric(leaf.fractal_order)

        score = TreeScoreTimewise()

        for layer_number in range(0, fm.number_of_layers + 1):
            simple_format = fm.get_simple_format(layer_number)

            v = simple_format.to_stream_voice(1)
            v.add_to_score(score, layer_number + 1)

        xml_path = path + '_test_7.xml'
        score.accidental_mode = 'normal'
        score.write(xml_path)
        self.assertCompareFiles(actual_file_path=xml_path)

    def test_8(self):
        fm = FractalMusic(proportions=[1, 2, 3, 4, 5, 6, 7], tree_permutation_order=[3, 6, 1, 5, 7, 2, 4])
        fm.tempo = 60
        fm.quarter_duration = 350
        fm.midi_generator.midi_range = [60, 72]
        fm.midi_generator.microtone = 4
        fm.tree_directions = [1, 1, -1, -1]

        fm.add_layer()
        fm.round_leaves()

        partial = fm.get_children()[6]
        partial.add_layer()
        partial.reduce_children(lambda ch: ch.fractal_order > partial.fractal_order)
        partial.round_leaves()

        result = [4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4]
        self.assertEqual(result, [node.midi_generator.microtone for node in fm.traverse()])

    def test_9(self):
        fm = FractalMusic(proportions=[1, 2, 3, 4, 5, 6, 7], tree_permutation_order=[3, 6, 2, 5, 1, 7, 4])

        fm.multi = (7, 4)
        fm.tempo = 60
        fm.quarter_duration = 70

        fm.midi_generator.microtone = 4
        # fm.quantize_leaves(0.5)
        fm.midi_generator.directions = [1, -1, 1, -1, 1, -1, 1]

        fm.midi_generator.midi_range = [36, 56]
        fm.add_layer()

        fm.reduce_children(lambda child: child.fractal_order > 4)
        fm.quantize_leaves(0.5)
        fm.get_children()[3].add_layer()
        fm.get_children()[3].reduce_children(lambda child: child.fractal_order > 4)
        for node in fm.traverse():
            node.chord.add_words(node.midi_value)
        text_path = path + '_test_9.txt'
        fm.write_infos(text_path)
        self.assertCompareFiles(actual_file_path=text_path)

        score = fm.get_score()
        xml_path = path + '_test_9.xml'
        score.write(xml_path)
        self.assertCompareFiles(actual_file_path=xml_path)

    def test_10(self):
        fm = FractalMusic(proportions=[1, 2, 3], tree_permutation_order=[3, 1, 2])
        fm.tempo = 60
        fm.duration = 10
        fm.midi_generator.midi_range = (60, 72)
        fm.permute_directions = True
        fm.midi_generator.set_directions(-1, 1, -1)

        fm.add_layer()

        for node in fm.traverse():
            node.chord.add_lyric(node.midi_generator.directions)
            node.chord.add_words(node.children_generated_midis, relative_y=30)
            node.chord.add_words(node.midi_generator.midi_range, relative_y=60)

        fm.add_layer()

        score = TreeScoreTimewise()
        score.accidental_mode = 'modern'
        score = fm.get_score(score)
        score.page_style.staff_distance = 150
        xml_path = path + '_test_10.xml'
        score.write(xml_path)
        self.assertCompareFiles(xml_path)



import os
from unittest import TestCase

from musurgia.pdf.pdf import Pdf
from musurgia.pdf.table import Table

path = os.path.abspath(__file__).split('.')[0]


# todo
class Test(TestCase):
    def test_1(self):
        pass
        # pdf_path = path + '_test_1.pdf'
        # pdf = Pdf()
        # table = Table(number_of_rows=20, relative_x=70)
        # table.draw(pdf=pdf)
        # pdf.write(pdf_path)



import itertools

from quicktions import Fraction

from musurgia.arithmeticprogression import ArithmeticProgression
from musurgia.pdf.line import HorizontalLineSegment
from musurgia.pdf.newdrawobject import DrawObject
from musurgia.pdf.rowcolumn import DrawObjectColumn, DrawObjectRow
from musurgia.permutation import LimitedPermutation, permute
from musurgia.tree import Tree


class FractalTreeException(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class ValueCannotBeChanged(FractalTreeException):
    def __init__(self, *args):
        super().__init__(*args)


class SetValueFirst(FractalTreeException):
    def __init__(self, *args):
        super().__init__('FractalTree().value must be set before add_layer()', *args)


class _Graphic(DrawObject):
    def __init__(self, fractal_tree, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._draw_object_column = None
        self._fractal_tree = fractal_tree
        self._factor = 1
        self._large_mark_line_max_length = 6
        self._large_mark_line_min_length = 3

    def update_draw_object_columns(self):
        factor = self.factor
        max_large_ml = self.large_mark_line_max_length
        min_large_ml = self.large_mark_line_min_length

        def _make_segment():
            segment = HorizontalLineSegment(length=node.value * factor)
            ml_length = (max_large_ml - (
                    node.get_distance() * (max_large_ml - min_large_ml) / self._fractal_tree.number_of_layers)) / 2
            segment.start_mark_line.length = ml_length
            segment.end_mark_line.length = ml_length
            return segment

        for node in self._fractal_tree.traverse():
            node.graphic._draw_object_column = DrawObjectColumn()
            node.graphic._draw_object_column.add_draw_object(_make_segment())

            if node.get_children():
                node.graphic._draw_object_column.add_draw_object(DrawObjectRow(top_margin=3))
            if node.up:
                node.up.graphic._draw_object_column.draw_objects[1].add_draw_object(node.graphic)
            if not node.up or node.up.get_children().index(node) == len(node.up.get_children()) - 1:
                node.graphic._draw_object_column.draw_objects[0].end_mark_line.show = True
                node.graphic._draw_object_column.draw_objects[0].end_mark_line.length *= 2
            if not node.up or node.up.get_children().index(node) == 0:
                node.graphic._draw_object_column.draw_objects[0].start_mark_line.length *= 2

    @property
    def factor(self):
        return self._factor

    @factor.setter
    def factor(self, val):
        self._factor = val

    @property
    def large_mark_line_max_length(self):
        return self._large_mark_line_max_length

    @large_mark_line_max_length.setter
    def large_mark_line_max_length(self, val):
        self._large_mark_line_max_length = val

    @property
    def large_mark_line_min_length(self):
        return self._large_mark_line_min_length

    @large_mark_line_min_length.setter
    def large_mark_line_min_length(self, val):
        self._large_mark_line_min_length = val

    @property
    def draw_object_column(self):
        if self._draw_object_column is None:
            self.update_draw_object_columns()
        return self._draw_object_column

    def add_labels(self, function, **kwargs):
        for node in self._fractal_tree.traverse():
            node.graphic.get_start_mark_line().add_label(function(node), **kwargs)

    def get_start_mark_line(self):
        return self.draw_object_column.draw_objects[0].start_mark_line

    def get_end_mark_line(self):
        return self.draw_object_column.draw_objects[0].end_mark_line

    def get_relative_x2(self):
        return self.draw_object_column.get_relative_x2()

    def get_relative_y2(self):
        return self.draw_object_column.get_relative_y2()

    def get_line_segment(self):
        return self.draw_object_column.draw_objects[0]

    def get_children_draw_object_row(self):
        return self.draw_object_column.draw_objects[1]

    def change_segment_attributes(self, **kwargs):
        for node in self._fractal_tree.traverse():
            for key in kwargs:
                setattr(node.graphic.get_line_segment(), key, kwargs[key])

    def draw(self, pdf):
        self.draw_object_column.draw(pdf)


class FractalTree(Tree):
    def __init__(self, value=None, proportions=None, tree_permutation_order=None, multi=None, fertile=True,
                 reading_direction='horizontal', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = None
        self._proportions = None
        self._tree_permutation_order = None
        self._multi = None
        self._permutation_order = None
        self._fertile = None
        self._reading_direction = None
        self._fractal_order = None
        self._children_fractal_values = None
        self._name = None
        self._graphic = _Graphic(self)

        self.proportions = proportions
        self.tree_permutation_order = tree_permutation_order
        self.value = value
        self.multi = multi
        self.fertile = fertile
        self.reading_direction = reading_direction

    # // private methods

    def _calculate_children_fractal_orders(self):
        if self.value and self.proportions:
            children_fractal_orders = range(1, self.size + 1)
            permutation_order = self.permutation_order
            if permutation_order:
                children_fractal_orders = permute(children_fractal_orders, permutation_order)
            return children_fractal_orders

    def _calculate_children_fractal_values(self):
        value = self.value
        if value and self.proportions:
            children_fractal_values = [value * prop for prop in self.proportions]
            if self.permutation_order:
                try:
                    children_fractal_values = permute(children_fractal_values, self.permutation_order)
                except ValueError:
                    raise ValueError('proportions and tree_permutation_order should have the same length')
            return children_fractal_values

    def _calculate_position_in_tree(self):
        parent = self.up
        if self.is_root:
            return 0
        else:
            index = parent.get_children().index(self)
            if index == 0:
                return parent.position_in_tree
            else:
                return parent.get_children()[index - 1].position_in_tree + parent.get_children()[index - 1].value

    def _change_children_value(self, factor):
        for child in self.get_children():
            child._value *= factor
            child._change_children_value(factor)

    def _check_merge_nodes(self, nodes):
        return True

    def _child_multi(self, parent, index):
        number_of_children = self.size
        multi_first = sum(parent.multi) % number_of_children
        if multi_first == 0:
            multi_first = number_of_children
        ch_multi = (multi_first, index + 1)
        return ch_multi

    def _get_merge_lengths(self, number_of_children, merge_index):
        if 0 > merge_index > self.size - 1:
            raise ValueError('generate_children.merge_index {} not valid'.format(merge_index))
        if number_of_children > self.size or number_of_children < 0:
            raise ValueError(
                'generate_children.number_of_children {} must be a positive int'.format(number_of_children))
        if number_of_children == 1:
            return [self.size]

        lengths = self.size * [1]
        pointer = merge_index
        sliced_lengths = [lengths[:pointer], lengths[pointer:]]

        if not sliced_lengths[0]:
            sliced_lengths = sliced_lengths[1:]

        while len(sliced_lengths) < number_of_children and len(sliced_lengths[0]) > 1:
            temp = sliced_lengths[0]
            sliced_lengths[0] = temp[:-1]
            sliced_lengths.insert(1, temp[-1:])

        while len(sliced_lengths) < number_of_children and len(sliced_lengths[pointer]) > 1:
            temp = sliced_lengths[pointer]
            sliced_lengths[pointer] = temp[:-1]
            sliced_lengths.insert(pointer + 1, temp[-1:])

        sliced_lengths = [len(x) for x in sliced_lengths]

        return sliced_lengths

    def _reduce(self):
        sisters = self.get_siblings()

        if not sisters:
            raise Exception('no reduction possible without sisters')
        else:

            non_zero_sisters = [sister for sister in sisters if sister.value != 0]

            if non_zero_sisters:
                addition = self.value / len(non_zero_sisters)
                for sister in non_zero_sisters:
                    if sister.value != 0:
                        sister._value += addition
                self._value = 0

    # // properties
    @property
    def children_fractal_values(self):
        if not self._children_fractal_values:
            self._children_fractal_values = self._calculate_children_fractal_values()
        return self._children_fractal_values

    @property
    def children_fractal_orders(self):
        return self._calculate_children_fractal_orders()

    @property
    def fertile(self):
        return self._fertile

    @fertile.setter
    def fertile(self, val):
        if not isinstance(val, bool):
            raise TypeError('fertile.value must be of type bool not{}'.format(type(val)))
        self._fertile = val

    @property
    def fractal_order(self):
        return self._fractal_order

    @property
    def layers(self):
        return [self.get_layer(i) for i in range(self.number_of_layers + 1)]

    @property
    def multi(self):
        return self._multi

    @multi.setter
    def multi(self, vals):
        if vals is None:
            vals = (1, 1)

        elif not isinstance(vals, tuple) or len(vals) != 2:
            raise TypeError('multi has to be a tuple with a length of 2')

        m_1 = vals[0]
        m_2 = vals[1]
        m_1, m_2 = (((m_1 - 1) % self.size) + 1 + ((m_2 - 1) // self.size)) % self.size, (
                (m_2 - 1) % self.size) + 1
        if m_1 == 0:
            m_1 = self.size
        self._multi = (m_1, m_2)

        self._permutation_order = None
        self._children_fractal_values = None

    @property
    def name(self):
        if self._name is None:
            if self.is_root:
                self._name = '0'
            elif self.up.is_root:
                self._name = str(self.up.get_children().index(self) + 1)
            else:
                self._name = self.up.name + '.' + str(self.up.get_children().index(self) + 1)
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def next(self):
        if not self.is_leaf():
            raise Exception('FractalTree().next property can only be used for leaves')
        else:
            try:

                parent = self.up
                while parent is not None:
                    parent_leaves = list(parent.traverse_leaves())
                    index = parent_leaves.index(self)
                    try:
                        return parent_leaves[index + 1]
                    except IndexError:
                        parent = parent.up
                raise IndexError()
            except IndexError:
                return None

    @property
    def number_of_layers(self):
        if self.get_leaves() == [self]:
            return 0
        else:
            return self.get_farthest_leaf().get_distance(self)

    @property
    def graphic(self):
        # if self._graphic is None:
        #     self._update_pdf_columns()
        return self._graphic

    @property
    def permutation_order(self):
        def _calculate_permutation_order():
            if self.tree_permutation_order:
                permutation = LimitedPermutation(input_list=list(range(1, self.size + 1)),
                                                 main_permutation_order=self.tree_permutation_order, multi=self.multi,
                                                 reading_direction=self.reading_direction)

                return permutation.permutation_order
            else:
                raise FractalTreeException('tree_permutation_order must be set')

        if self._permutation_order is None:
            self._permutation_order = _calculate_permutation_order()
        return self._permutation_order

    @property
    def position(self):
        if self.is_root:
            # return self.first_position
            return 0
        else:
            siblings = self.up.get_children()
            index = siblings.index(self)
            previous_siblings = siblings[:index]
            position_in_parent = 0
            for child in previous_siblings: position_in_parent += child.value
            return position_in_parent + self.up.position

    @property
    def position_in_tree(self):
        return self._calculate_position_in_tree()

    @property
    def proportions(self):
        return self._proportions

    @proportions.setter
    def proportions(self, values):
        if values is None:
            values = [1, 2, 3]
        self._proportions = [Fraction(Fraction(value) / Fraction(sum(values))) for value in values]

    @property
    def reading_direction(self):
        return self._reading_direction

    @reading_direction.setter
    def reading_direction(self, val):
        if self._reading_direction:
            raise FractalTreeException('reading_direction can only be set during initialisation')
        permitted = ['horizontal', 'vertical']
        if val not in permitted:
            raise ValueError('reading_direction.value {} must be in {}'.format(val, permitted))
        self._reading_direction = val

    @property
    def size(self):
        return len(self.proportions)

    @property
    def tree_permutation_order(self):
        return self._tree_permutation_order

    @tree_permutation_order.setter
    def tree_permutation_order(self, values):
        if values is None:
            values = [3, 1, 2]
        self._tree_permutation_order = values
        self._permutation_order = None

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, val):
        if val:
            val = Fraction(val)
        if self._value:
            if self.is_root and not self.get_children():
                self._value = val
            else:
                raise ValueCannotBeChanged(
                    'FractalTree().value can only be changed for the root with no children. Use change_value() for other cases.'
                )
        else:
            self._value = val

        if self._value == 0:
            self.fertile = False

    # //public methods
    # add
    def add_layer(self, *conditions):
        if self.value is None:
            raise SetValueFirst()

        leaves = list(self.traverse_leaves())
        if not leaves:
            leaves = [self]

        if conditions:
            for leaf in leaves:
                for condition in conditions:
                    if not condition(leaf):
                        leaf.fertile = False
                        break

        for leaf in leaves:
            if leaf.fertile is True:
                for i in range(leaf.size):
                    new_node = leaf.copy()
                    new_node.value = leaf.children_fractal_values[i]
                    new_node.multi = self._child_multi(leaf, i)
                    leaf.add_child(new_node)
                    new_node._fractal_order = leaf.children_fractal_orders[i]

            else:
                pass

    def add_self(self):
        leaves = self.get_leaves()
        for leaf in leaves:
            new_node = self.copy()
            new_node._up = self
            new_node._fractal_order = self.fractal_order
            leaf.add_child(new_node)

    # get
    def get_layer(self, layer, key=None):
        if layer <= self.get_root().number_of_layers:
            branch_distances = []
            for child in self.get_children():
                branch_distances.append(child.get_farthest_leaf().get_distance())
            if layer == 0:
                if key:
                    return [getattr(self, key)]
                else:
                    return [self]

            if layer >= 1:
                output = []
                if not branch_distances:
                    output.extend(self.get_layer(layer=layer - 1, key=key))
                for i in range(len(self.get_children())):
                    child = self.get_children()[i]
                    if branch_distances[i] == 1:
                        if key:
                            output.append(getattr(child, key))
                        else:
                            output.append(child)
                    else:
                        if layer == 1:
                            output.extend(child.get_layer(layer - 1, key))
                        else:
                            output.append(child.get_layer(layer - 1, key))

                return output
        else:
            err = 'max layer number=' + str(self.number_of_layers)
            raise ValueError(err)

    def get_next_sibling(self):
        try:
            siblings = self.up.get_children()
            index = siblings.index(self)
            return siblings[index + 1]
        except (IndexError, AttributeError):
            return None

    def get_previous_leaf(self):
        if not self.is_leaf():
            raise Exception('FractalTree().get_previous_leaf  can only be applied to leaves')
        try:
            parent = self.up
            while parent is not None:
                parent_leaves = list(parent.traverse_leaves())
                index = parent_leaves.index(self)
                try:
                    return parent_leaves[index - 1]
                except IndexError:
                    parent = parent.up
            raise IndexError()
        except IndexError:
            return None

    def get_previous_sibling(self):
        try:
            siblings = self.up.get_children()
            index = siblings.index(self)
            if index == 0:
                return None
            return siblings[index - 1]
        except AttributeError:
            return None

    # other

    def change_value(self, new_value):
        if not new_value:
            new_value = Fraction(new_value)
        if not self.value:
            self.value = new_value
        else:
            factor = Fraction(new_value, self.value)
            self._value = new_value
            for node in reversed(self.get_branch()[:-1]):
                node._value = sum([child.value for child in node.get_children()])

            self._change_children_value(factor)

    def generate_children(self, number_of_children, mode='reduce', merge_index=0):

        permitted_modes = ['reduce', 'reduce_backwards', 'reduce_forwards', 'reduce_sieve', 'merge']
        if mode not in permitted_modes:
            raise ValueError('generate_children.mode {} must be in{}'.format(mode, permitted_modes))
        if isinstance(number_of_children, int):
            if number_of_children == 0:
                pass
            if number_of_children > self.size:
                raise ValueError(
                    'generate_children.number_of_children {} can not be a greater than size {}'.format(
                        number_of_children, self.size))
            if number_of_children < 0:
                raise ValueError(
                    'generate_children.number_of_children {} must be a positive int'.format(number_of_children))
            else:
                self.add_layer()
                if mode in ['reduce', 'reduce_backwards']:
                    self.reduce_children(
                        lambda child: child.fractal_order < self.size - number_of_children + 1)
                elif mode == 'reduce_forwards':
                    self.reduce_children(
                        lambda child: child.fractal_order > number_of_children)
                elif mode == 'reduce_sieve':
                    if number_of_children == 1:
                        self.reduce_children(condition=lambda child: child.fractal_order not in [1])
                    else:
                        ap = ArithmeticProgression(a1=1, an=self.size, n=number_of_children)
                        selection = [int(round(x)) for x in ap]
                        self.reduce_children(condition=lambda child: child.fractal_order not in selection)
                else:
                    merge_lengths = self._get_merge_lengths(number_of_children, merge_index)
                    self.merge_children(*merge_lengths)

        elif isinstance(number_of_children, tuple):
            self.generate_children(len(number_of_children), mode=mode, merge_index=merge_index)

            for index, child in enumerate(self.get_children()):
                if mode == 'reduce':
                    number_of_grand_children = number_of_children[
                        child.fractal_order - child.size + len(number_of_children) - 1]
                else:
                    number_of_grand_children = number_of_children[index]
                child.generate_children(number_of_grand_children, mode=mode, merge_index=merge_index)

        else:
            raise TypeError()

    def merge_children(self, *lengths):
        children = self.get_children()
        if not children:
            raise Exception('there are no children to be merged')
        if sum(lengths) != len(children):
            raise ValueError(
                'sum of lengths {} must be the same as length of children {}'.format(sum(lengths), len(children)))

        def _merge(nodes):
            self._check_merge_nodes(nodes)

            node_values = [node.value for node in nodes]

            nodes[0]._value = sum(node_values)
            for node in nodes[1:]:
                self.remove_child(node)

        iter_children = iter(children)
        chunks = [list(itertools.islice(iter_children, l)) for l in lengths]

        for chunk in chunks:
            _merge(chunk)

    def reduce_children(self, condition):

        for child in self.get_children():
            if condition(child):
                child._reduce()

        to_be_detached = [child for child in self.get_children() if condition(child)]
        for child in to_be_detached:
            child.detach()

        self._children_fractal_values = [child.value for child in self.get_children()]

    def split(self, *proportions):

        for prop in proportions:
            self.add_self()
            self.get_children()[-1].value = self.value * prop / sum(proportions)

    # // copy

    def copy(self):
        return self.__class__(value=self.value, proportions=self.proportions,
                              tree_permutation_order=self.tree_permutation_order, multi=self.multi,
                              reading_direction=self.reading_direction, fertile=self.fertile)

    # def copy_without_children(self):

    def __deepcopy__(self, memodict={}):
        copied = self.__class__(value=self.value, proportions=self.proportions,
                                tree_permutation_order=self.tree_permutation_order, multi=self.multi,
                                reading_direction=self.reading_direction, fertile=self.fertile)
        copied._fractal_order = self.fractal_order
        copied._name = self._name
        for child in self.get_children():
            copied.add_child(child.__deepcopy__())
        return copied



import os
from unittest import TestCase
from quicktions import Fraction

from musurgia.fractaltree.fractaltree import FractalTree

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):

    def test_1(self):
        ft = FractalTree(proportions=(1, 2, 3), tree_permutation_order=(3, 1, 2), value=10)
        ft.add_layer()
        ft.add_layer()
        for node in ft.get_layer(1):
            node.reduce_children(condition=lambda node: node.fractal_order == 1)
        self.assertEqual([node.fractal_order for node in ft.traverse_leaves()], [2, 3, 3, 2, 2, 3])

    def test_2(self):
        ft = FractalTree(proportions=(1, 2, 3), tree_permutation_order=(3, 1, 2), value=10)
        ft.add_layer()
        ft.add_layer()
        for node in ft.get_layer(1):
            node.reduce_children(condition=lambda node: node.fractal_order == 1)
        self.assertEqual([node.value for node in ft.traverse_leaves()],
                         [Fraction(25, 12), Fraction(35, 12), Fraction(35, 36), Fraction(25, 36), Fraction(25, 18),
                          Fraction(35, 18)])



from musurgia.pdf.masterslave import Master, PositionMaster, MarginMaster, Slave, PositionSlave, MarginSlave
from musurgia.pdf.newdrawobject import DrawObject
from musurgia.unittest import TestCase


class DummyMaster(Master):

    def get_slave_position(self, slave, position):
        pass

    def get_slave_margin(self, slave, margin):
        if margin in ['t', 'top']:
            return 10


class DummyPositionMaster(PositionMaster):
    def get_slave_position(self, slave, position):
        pass


class DummyMarginMaster(MarginMaster):
    def get_slave_margin(self, slave, margin):
        if margin in ['t', 'top']:
            return 20


class DummyMarginSlave(DrawObject, MarginSlave):
    def get_relative_y2(self):
        return 'dummy'

    def get_relative_x2(self):
        return 'dummy'

    def draw(self, pdf):
        return 'dummy'


class TestSlave(TestCase):
    def test_slave_init(self):
        s = Slave(name='test_slave', master=DummyMaster())
        self.assertTrue(isinstance(s, Slave))

    def test_position_slave_init(self):
        s = PositionSlave(name='test_position_slave', master=DummyPositionMaster())
        self.assertTrue(isinstance(s, PositionSlave))

    def test_margin_slave_init(self):
        s = MarginSlave(name='test_margin_slave', master=DummyMarginMaster())
        self.assertTrue(isinstance(s, MarginSlave))

    def test_margin_slave_top(self):
        m = DummyMarginMaster()
        s = MarginSlave(name='test_margin_slave', master=m)

        actual = s.top_margin
        expected = m.get_slave_margin(s, 't')
        self.assertTrue(expected, actual)

    def test_margin_slave_position(self):
        m = DummyMarginMaster()
        s = DummyMarginSlave(name='test_margin_slave', master=m)
        s.relative_x = 10



from unittest import TestCase
import os

from musurgia.fractaltree.fractaltree import FractalTree

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):

    def test_1(self):
        ft = FractalTree(reading_direction='vertical', value=10)
        ft.add_layer()
        permutation_orders = [child.permutation_order for child in ft.get_children()]
        result = [[2, 1, 3], [3, 2, 1], [1, 3, 2]]
        self.assertEqual(result, permutation_orders)



from abc import ABC, abstractmethod

from musurgia.pdf.margined import Margined
from musurgia.pdf.positioned import Positioned


class DrawObject(ABC, Positioned, Margined):
    def __init__(self, show=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._show = None
        self._page_break = False
        self._line_break = False
        self._parent = None
        self.show = show

    @property
    def parent(self):
        return self._parent

    def _check_line_break(self, pdf):
        next_x2 = pdf.x + self.get_relative_x2()
        printable_range = pdf.w - pdf.r_margin
        diff = next_x2 - printable_range
        if diff > 0:
            self._line_break = True
            self.relative_x -= printable_range
            if self.relative_x < 0:
                self.relative_x = 0

            if self.parent:
                bottom_margin = self.parent.bottom_margin
            else:
                bottom_margin = self.bottom_margin
            pdf.y += self.get_relative_y2() + bottom_margin
            # pdf.y += self.get_relative_y2()
            pdf.x = pdf.l_margin

    def _check_page_break(self, pdf):
        next_y2 = pdf.y + self.get_relative_y2() + self.bottom_margin
        printable_y_range = pdf.h - pdf.b_margin
        diff = next_y2 - printable_y_range
        if diff > 0:
            self._page_break = True
            # self.relative_y -= printable_y_range
            # if self.relative_y < 0:
            #     self.relative_y = 0

            margins = pdf.l_margin, pdf.t_margin, pdf.r_margin, pdf.b_margin
            pdf.add_page()
            pdf.l_margin, pdf.t_margin, pdf.r_margin, pdf.b_margin = margins
            # if self.parent:
            #     top_margin = self.parent.top_margin
            # else:
            #     top_margin = self.top_margin
            # pdf.y = pdf.t_margin + top_margin
            pdf.y = pdf.t_margin
            pdf.x = pdf.l_margin

    @abstractmethod
    def get_relative_x2(self):
        raise NotImplementedError()

    @abstractmethod
    def get_relative_y2(self):
        raise NotImplementedError()

    def get_height(self):
        return self.get_relative_y2() - self.relative_y

    @abstractmethod
    def draw(self, pdf):
        raise NotImplementedError()

    @property
    def show(self):
        return self._show

    @show.setter
    def show(self, val):
        if not isinstance(val, bool):
            raise TypeError()
        self._show = val

    def draw_with_break(self, pdf):
        self._check_line_break(pdf)
        self._check_page_break(pdf)
        self.draw(pdf)



from pathlib import Path

from musurgia.pdf.line import HorizontalSegmentedLine
from musurgia.pdf.pdf import Pdf
from musurgia.unittest import TestCase

path = Path(__file__)


class TestHorizontalSegmentedLine(TestCase):
    def setUp(self) -> None:
        self.pdf = Pdf()
        self.hsl = HorizontalSegmentedLine(lengths=[10, 15, 20, 25])

    def test_draw(self):
        with self.file_path(path, 'draw', 'pdf') as pdf_path:
            with self.file_path(parent_path=path, name='draw', extension='pdf') as pdf_path:
                self.pdf.translate_page_margins()
                self.pdf.draw_ruler('h')
                self.pdf.draw_ruler('v')
                self.pdf.translate(10, 10)
                self.hsl.draw(self.pdf)
                self.pdf.write(pdf_path)
